"""
Guesslang: a machine learning program that guesses the programming language
of a given source code.

"""

import os


# Do not let Tensorflow print its numerous warning messages on startup.
# Unless the user asked to see them by setting Tensorflow logging level.
if "TF_CPP_MIN_LOG_LEVEL" not in os.environ:
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"


from .guess import Guess, GuesslangError  # noqa: F401


__version__ = "2.2.2"



import os
import pandas as pd
from ekorpkit import eKonf
from ekorpkit.corpora import Corpus, Corpora
from ekorpkit.datasets import Dataset, Datasets
from ekorpkit.io.file import load_dataframe


class Data:
    def __init__(self, data=None, **args):

        self.args = args
        self._data = None
        self._column = None
        self.verbose = args.get("verbose", False)

        if isinstance(data, pd.DataFrame):
            self._data = data
        elif eKonf.is_config(args):
            self._load(**args)

        _data = eKonf.function(args, "concat_dataframes", data=self._data)
        if _data is not None:
            self._data = _data

        self._column_info = self.args.get("column_info")
        if self._column_info is None:
            raise ValueError("Column info can't be None")
        if self._column is None:
            self._column = eKonf.instantiate(self._column_info)

    def _load(self, **args):
        corpus = args.get(eKonf.Keys.CORPUS) or {}
        dataset = args.get(eKonf.Keys.DATASET) or {}
        data_dir = args.get("data_dir")
        data_file = args.get("data_file")

        if corpus.get("name") and dataset.get("name") is None:
            args = corpus
        elif dataset.get("name"):
            args = dataset

        if eKonf.is_instantiatable(args):
            data = eKonf.instantiate(args)
            if isinstance(data, (Corpus)):
                self._data = data.data
                self._column = data.COLUMN
            elif isinstance(data, (Corpora)):
                data.concat_corpora()
                self._data = data.data
                self._column = data.COLUMN
            elif isinstance(data, (Dataset, Datasets)):
                self._data = data.splits
                self._column = data.COLUMN
        elif data_dir and data_file:
            self._data = eKonf.function(args, "load_dataframe")

    def __str__(self):
        classname = self.__class__.__name__
        return f"{classname}:\n{self.args}"

    @property
    def COLUMN(self):
        return self._column

    @property
    def ID(self):
        return self.COLUMN.ID

    @property
    def IDs(self):
        return self.COLUMN.IDs

    @property
    def TEXT(self):
        return self.COLUMN.TEXT

    @property
    def data(self):
        return self._data



from pathlib import Path
from ekorpkit import eKonf
from hydra.utils import instantiate

from ..datasets import ShardingDataset
from ..models import train_tokenizer
from ..utils.func import elapsed_timer


def lmdata(**args):
    args = eKonf.to_config(args)

    with elapsed_timer(format_time=True) as elapsed:
        for subtask, task_args in args.task.lmdata.items():
            if "_target_" in task_args:
                instantiate(task_args, parent=args, _recursive_=False)
            else:
                print(f"{subtask} is not a valid subtask")
        print(f">>> Elapsed time: {elapsed()} <<< ")


def build_vocab(**args):
    cfg = eKonf.to_config(args)
    # print(cfg)
    args = cfg.parent.tokenizer

    if isinstance(args.input_dir, str):
        input_dirs = [args.input_dir]
    else:
        input_dirs = list(args.input_dir)

    if args.train_files is not None:
        # train_files = {args.input_dir + '/' + f:d for f, d in args.train_files.items()}
        train_files = []
        for input_dir in input_dirs:
            train_files = [input_dir + "/" + f for f in args.train_files]
    else:
        # train_files = {str(f):1 for f in Path(args.input_dir).glob('**/*') if f.is_file()}
        train_files = []
        for input_dir in input_dirs:
            train_files += [str(f) for f in Path(input_dir).glob("**/*") if f.is_file()]
    train_tokenizer(train_files=train_files, args=args, output_dir=args.output_dir)


def sharding(**args):
    args = eKonf.to_config(args)
    # print(cfg)

    sharding = ShardingDataset(**args)
    sharding.load_datasets()
    sharding.segment_articles_into_sentences()
    sharding.distribute_datasets_over_shards()
    sharding.write_shards_to_disk()



import logging
from ekorpkit import eKonf
from ekorpkit.pipelines.pipe import apply_pipe


log = logging.getLogger(__name__)


def eval_classification(data=None, **args):
    args = eKonf.to_config(args)
    to_eval = args.to_eval
    pipe = args.pipeline.load_dataframe
    eval_metrics = args[eKonf.Keys.FUNC]
    if data is None:
        data = apply_pipe(data, pipe)
    eval_metrics = eKonf.partial(eval_metrics)
    cm = eval_metrics(data[to_eval.actual], data[to_eval.predicted])
    plot = args.visualize.plot
    eKonf.instantiate(plot, data=cm)



import logging
import os
import random
import hydra
import pathlib
import dotenv
import ekorpkit.utils.batch.batcher as batcher
from enum import Enum
from hydra.core.config_store import ConfigStore
from omegaconf import OmegaConf, SCMode, DictConfig, ListConfig
from typing import Any, List, IO, Dict, Union, Tuple, Optional
from ekorpkit.io.cached_path import cached_path
from ekorpkit.utils.func import lower_case_with_underscores
from . import _version


log = logging.getLogger(__name__)

__hydra_version_base__ = "1.2"


def __ekorpkit_path__():
    return pathlib.Path(__file__).parent.as_posix()


def __home_path__():
    return pathlib.Path.home().as_posix()


def __version__():
    return _version.get_versions()["version"]


def check_path(path: str, alt_path: str = None):
    if os.path.exists(path):
        return path
    elif alt_path:
        return alt_path


def _today(_format="%Y-%m-%d"):
    from datetime import datetime

    if _format is None:
        return datetime.today().date()
    else:
        return datetime.today().strftime(_format)


def _now(_format="%Y-%m-%d %H:%M:%S"):
    from datetime import datetime

    if _format is None:
        return datetime.now()
    else:
        return datetime.now().strftime(_format)


def _strptime(
    _date_str: str,
    _format: str = "%Y-%m-%d",
):
    from datetime import datetime

    return datetime.strptime(_date_str, _format)


def _to_dateparm(_date, _format="%Y-%m-%d"):
    from datetime import datetime

    _dtstr = datetime.strftime(_date, _format)
    _dtstr = "${to_datetime:" + _dtstr + "," + _format + "}"
    return _dtstr


def _path(
    url_or_filename,
    extract_archive: bool = False,
    force_extract: bool = False,
    return_dir: bool = False,
    cache_dir=None,
    verbose: bool = False,
):
    return cached_path(
        url_or_filename,
        extract_archive=extract_archive,
        force_extract=force_extract,
        return_dir=return_dir,
        cache_dir=cache_dir,
        verbose=verbose,
    )


def _compose(
    overrides: List[str] = [],
    config_group: str = None,
    *,
    return_as_dict: bool = False,
    throw_on_resolution_failure: bool = True,
    throw_on_missing: bool = False,
    config_name="ekonf",
    verbose: bool = False,
):
    """
    Compose your configuration from config groups and overrides (overrides=["override_name"])

    :param overrides: List of overrides to apply
    :param config_group: Config group name to select ('config_group=name')
    :param return_as_dict: Return the composed config as a dict
    :param throw_on_resolution_failure: Throw if resolution fails
    :param throw_on_missing: Throw if a config is missing
    :param config_name: Name of the config to compose
    :param verbose: Print the composed config

    :return: The composed config
    """
    if config_group:
        _task = config_group.split("=")
        if len(_task) == 2:
            key, value = _task
        else:
            key = _task[0]
            value = "default"
        config_group = f"{key}={value}"
    else:
        key = None
        value = None
    if key and value:
        with hydra.initialize_config_module(
            config_module="ekorpkit.conf", version_base=__hydra_version_base__
        ):
            cfg = hydra.compose(config_name=config_name, overrides=overrides)
            cfg = _select(
                cfg,
                key=key,
                default=None,
                throw_on_missing=False,
                throw_on_resolution_failure=False,
            )
            if cfg:
                overide = config_group
            else:
                overide = f"+{config_group}"
            if overrides:
                overrides.append(overide)
            else:
                overrides = [overide]
    if verbose:
        print(f"compose config with overrides: {overrides}")
    with hydra.initialize_config_module(
        config_module="ekorpkit.conf", version_base=__hydra_version_base__
    ):
        cfg = hydra.compose(config_name=config_name, overrides=overrides)
        if key:
            cfg = _select(
                cfg,
                key=key,
                default=None,
                throw_on_missing=throw_on_missing,
                throw_on_resolution_failure=throw_on_resolution_failure,
            )
        if verbose:
            print(cfg)
        if return_as_dict and isinstance(cfg, DictConfig):
            return _to_dict(cfg)
        return cfg


def _to_dict(
    cfg: Any,
):
    if isinstance(cfg, dict):
        cfg = _to_config(cfg)
    if isinstance(cfg, (DictConfig, ListConfig)):
        return OmegaConf.to_container(
            cfg,
            resolve=True,
            throw_on_missing=False,
            structured_config_mode=SCMode.DICT,
        )
    return cfg


def _to_config(
    cfg: Any,
):
    return OmegaConf.create(cfg)


def dotenv_values(dotenv_path=None, **kwargs):
    config = dotenv.dotenv_values(dotenv_path=dotenv_path, **kwargs)
    return dict(config)


def getcwd():
    try:
        return hydra.utils.get_original_cwd()
    except:
        return os.getcwd()


_env_initialized_ = False

_config = _compose().copy()

DictKeyType = Union[str, int, Enum, float, bool]

OmegaConf.register_new_resolver("__ekorpkit_path__", __ekorpkit_path__)
OmegaConf.register_new_resolver("__home_path__", __home_path__)
OmegaConf.register_new_resolver("__version__", __version__)
OmegaConf.register_new_resolver("today", _today)
OmegaConf.register_new_resolver("to_datetime", _strptime)
OmegaConf.register_new_resolver("iif", lambda cond, t, f: t if cond else f)
OmegaConf.register_new_resolver("randint", random.randint, use_cache=True)
OmegaConf.register_new_resolver("get_method", hydra.utils.get_method)
OmegaConf.register_new_resolver("get_original_cwd", getcwd)
OmegaConf.register_new_resolver("check_path", check_path)
OmegaConf.register_new_resolver("cached_path", _path)
OmegaConf.register_new_resolver(
    "lower_case_with_underscores", lower_case_with_underscores
)
OmegaConf.register_new_resolver("dotenv_values", dotenv_values)


class _Keys(str, Enum):
    """Special keys in configs used by instantiate."""

    TARGET = "_target_"
    CONVERT = "_convert_"
    RECURSIVE = "_recursive_"
    ARGS = "_args_"
    PARTIAL = "_partial_"
    CONFIG = "_config_"
    CONFIG_GROUP = "_config_group_"
    PIPELINE = "_pipeline_"
    TASK = "_task_"
    CALL = "_call_"
    EXEC = "_exec_"
    PARMS = "_parms_"
    METHOD = "_method_"
    FUNC = "_func_"
    NAME = "_name_"
    SPLIT = "split"
    CORPUS = "corpus"
    DATASET = "dataset"
    ID = "id"
    TEXT = "text"
    TIMESTAMP = "timestamp"
    DATETIME = "datetime"


def _methods(cfg: Any, obj: object):
    cfg = eKonf.to_dict(cfg)
    if not cfg:
        log.info("No method defined to call")
        return

    if isinstance(cfg, dict) and _Keys.METHOD in cfg:
        _method_ = cfg[_Keys.METHOD]
    else:
        _method_ = cfg
    if isinstance(_method_, str):
        log.info(f"Calling {_method_}")
        return getattr(obj, _method_)()
    elif isinstance(_method_, dict):
        log.info(f"Calling {_method_}")
        if _Keys.CALL in _method_:
            _call_ = _method_.pop(_Keys.CALL)
        else:
            _call_ = True
        if _call_:
            return getattr(obj, _method_[_Keys.NAME])(**_method_[_Keys.PARMS])
        else:
            log.info(f"Skipping call to {_method_}")
    elif isinstance(_method_, list):
        for _each_method in _method_:
            log.info(f"Calling {_each_method}")
            if isinstance(_each_method, str):
                getattr(obj, _each_method)()
            elif isinstance(_each_method, dict):
                if _Keys.CALL in _each_method:
                    _call_ = _each_method.pop(_Keys.CALL)
                else:
                    _call_ = True
                if _call_:
                    getattr(obj, _each_method[_Keys.NAME])(**_each_method[_Keys.PARMS])
                else:
                    log.info(f"Skipping call to {_each_method}")


def _function(cfg: Any, _name_, return_function=False, **parms):
    cfg = eKonf.to_dict(cfg)
    if not isinstance(cfg, dict):
        log.info("No function defined to execute")
        return None

    if _Keys.FUNC not in cfg:
        log.info("No function defined to execute")
        return None

    _functions_ = cfg[_Keys.FUNC]
    fn = _partial(_functions_[_name_])
    if _name_ in cfg:
        _parms = cfg[_name_]
        _parms = {**_parms, **parms}
    else:
        _parms = parms
    if _Keys.EXEC in _parms:
        _exec_ = _parms.pop(_Keys.EXEC)
    else:
        _exec_ = True
    if _exec_:
        if callable(fn):
            if return_function:
                log.info(f"Returning function {fn}")
                return fn
            log.info(f"Executing function {fn} with parms {_parms}")
            return fn(**_parms)
        else:
            log.info(f"Function {_name_} not callable")
            return None
    else:
        log.info(f"Skipping execute of {fn}")
        return None


def _print(cfg: Any, resolve: bool = True, **kwargs):
    import pprint

    if _is_config(cfg):
        if resolve:
            pprint.pprint(_to_dict(cfg), **kwargs)
        else:
            pprint.pprint(cfg, **kwargs)
    else:
        print(cfg)


def _select(
    cfg: Any,
    key: str,
    *,
    default: Any = None,
    throw_on_resolution_failure: bool = True,
    throw_on_missing: bool = False,
):
    key = key.replace("/", ".")
    return OmegaConf.select(
        cfg,
        key=key,
        default=default,
        throw_on_resolution_failure=throw_on_resolution_failure,
        throw_on_missing=throw_on_missing,
    )


def _is_config(
    cfg: Any,
):
    return isinstance(cfg, (DictConfig, dict))


def _is_instantiatable(cfg: Any):
    return _is_config(cfg) and _Keys.TARGET in cfg


def _load(file_: Union[str, pathlib.Path, IO[Any]]) -> Union[DictConfig, ListConfig]:
    return OmegaConf.load(file_)


def _save(
    config: Any, f: Union[str, pathlib.Path, IO[Any]], resolve: bool = False
) -> None:
    OmegaConf.save(config, f, resolve=resolve)


def _update(_dict, _overrides):
    import collections.abc

    for k, v in _overrides.items():
        if isinstance(v, collections.abc.Mapping):
            _dict[k] = _update((_dict.get(k) or {}), v)
        else:
            _dict[k] = v
    return _dict


def _merge(
    *configs: Union[
        DictConfig,
        ListConfig,
        Dict[DictKeyType, Any],
        List[Any],
        Tuple[Any, ...],
        Any,
    ],
) -> Union[ListConfig, DictConfig]:
    """
    Merge a list of previously created configs into a single one
    :param configs: Input configs
    :return: the merged config object.
    """
    return OmegaConf.merge(*configs)


def _to_yaml(cfg: Any, *, resolve: bool = False, sort_keys: bool = False) -> str:
    return OmegaConf.to_yaml(cfg, resolve=resolve, sort_keys=sort_keys)


def _to_container(
    cfg: Any,
    *,
    resolve: bool = False,
    throw_on_missing: bool = False,
    enum_to_str: bool = False,
    structured_config_mode: SCMode = SCMode.DICT,
):
    return OmegaConf.to_container(
        cfg,
        resolve=resolve,
        throw_on_missing=throw_on_missing,
        enum_to_str=enum_to_str,
        structured_config_mode=structured_config_mode,
    )


def _run(config: Any, **kwargs: Any) -> Any:
    config = _merge(config, kwargs)
    _config_ = config.get(_Keys.CONFIG)
    if _config_ is None:
        log.warning("No _config_ specified in config")
        return None
    if isinstance(_config_, str):
        _config_ = [_config_]
    for _cfg_ in _config_:
        cfg = _select(config, _cfg_)
        _instantiate(cfg)


def _partial(
    config: Any = None, config_group: str = None, *args: Any, **kwargs: Any
) -> Any:
    if config is None and config_group is None:
        log.warning("No config specified")
        return None
    elif config_group is not None:
        config = _compose(config_group=config_group)
    kwargs[_Keys.PARTIAL] = True
    return _instantiate(config, *args, **kwargs)


def _instantiate(config: Any, *args: Any, **kwargs: Any) -> Any:
    """
    :param config: An config object describing what to call and what params to use.
                   In addition to the parameters, the config must contain:
                   _target_ : target class or callable name (str)
                   And may contain:
                   _args_: List-like of positional arguments to pass to the target
                   _recursive_: Construct nested objects as well (bool).
                                False by default.
                                may be overridden via a _recursive_ key in
                                the kwargs
                   _convert_: Conversion strategy
                        none    : Passed objects are DictConfig and ListConfig, default
                        partial : Passed objects are converted to dict and list, with
                                  the exception of Structured Configs (and their fields).
                        all     : Passed objects are dicts, lists and primitives without
                                  a trace of OmegaConf containers
                   _partial_: If True, return functools.partial wrapped method or object
                              False by default. Configure per target.
                   _args_: List-like of positional arguments
    :param args: Optional positional parameters pass-through
    :param kwargs: Optional named parameters to override
                   parameters in the config object. Parameters not present
                   in the config objects are being passed as is to the target.
                   IMPORTANT: dataclasses instances in kwargs are interpreted as config
                              and cannot be used as passthrough
    :return: if _target_ is a class name: the instantiated object
             if _target_ is a callable: the return value of the call
    """
    if not _env_initialized_:
        _init_env_()
    if not _is_instantiatable(config):
        log.warning(f"Config is not instantiatable, returning config")
        return config
    _recursive_ = config.get(_Keys.RECURSIVE, False)
    if _Keys.RECURSIVE not in kwargs:
        kwargs[_Keys.RECURSIVE] = _recursive_
    return hydra.utils.instantiate(config, *args, **kwargs)


def _load_dotenv(verbose=False):
    original_cwd = getcwd()
    dotenv_path = pathlib.Path(original_cwd, ".env")
    dotenv.load_dotenv(dotenv_path=dotenv_path, verbose=verbose)


def _init_env_(cfg=None, verbose=False):
    global _env_initialized_

    _load_dotenv(verbose=verbose)

    if cfg is None:
        cfg = _config
    env = cfg.env

    backend = env.distributed_framework.backend
    for env_name, env_value in env.get("os", {}).items():
        if env_value:
            if verbose:
                log.info(f"setting environment variable {env_name} to {env_value}")
            os.environ[env_name] = str(env_value)

    if env.distributed_framework.initialize:
        backend_handle = None
        if backend == "ray":
            import ray

            ray_cfg = env.get("ray", None)
            ray_cfg = eKonf.to_container(ray_cfg, resolve=True)
            if verbose:
                log.info(f"initializing ray with {ray_cfg}")
            ray.init(**ray_cfg)
            backend_handle = ray

        elif backend == "dask":
            from dask.distributed import Client

            dask_cfg = env.get("dask", None)
            dask_cfg = eKonf.to_container(dask_cfg, resolve=True)
            if verbose:
                log.info(f"initializing dask client with {dask_cfg}")
            client = Client(**dask_cfg)
            if verbose:
                log.info(client)

        batcher.batcher_instance = batcher.Batcher(
            backend_handle=backend_handle, **env.batcher
        )
        if verbose:
            log.info(batcher.batcher_instance)
    _env_initialized_ = True


def _stop_env_(cfg, verbose=False):
    env = cfg.env
    backend = env.distributed_framework.backend

    if env.distributed_framework.initialize:
        if backend == "ray":
            import ray

            if ray.is_initialized():
                ray.shutdown()
                if verbose:
                    log.info("shutting down ray")

        # elif modin_engine == 'dask':
        #     from dask.distributed import Client

        #     if Client.initialized():
        #         client.close()
        #         log.info(f'shutting down dask client')


def apply_pipe(df, pipe):
    _func_ = pipe.get(_Keys.FUNC)
    fn = eKonf.partial(_func_)
    log.info(f"Applying pipe: {fn}")
    if isinstance(df, dict):
        if "concat_dataframes" in str(fn):
            return fn(df, pipe)
        else:
            dfs = {}
            for df_no, df_name in enumerate(df):
                df_each = df[df_name]
                log.info(
                    f"Applying pipe to dataframe [{df_name}], {(df_no+1)}/{len(df)}"
                )
                pipe[_Keys.NAME] = df_name
                dfs[df_name] = fn(df_each, pipe)
            return dfs
    else:
        return fn(df, pipe)


def _dependencies(key, path=None):
    import re
    from collections import defaultdict

    if path is None:
        path = os.path.join(
            os.path.dirname(__file__), "resources", "requirements-extra.txt"
        )

    with open(path) as fp:
        extra_deps = defaultdict(set)
        for k in fp:
            if k.strip() and not k.startswith("#"):
                tags = set()
                if ":" in k:
                    k, v = k.split(":")
                    tags.update(vv.strip() for vv in v.split(","))
                tags.add(re.split("[<=>]", k.strip())[0])
                for t in tags:
                    extra_deps[t].add(k.strip())

        # add tag `exhaustive` at the end
        extra_deps["exhaustive"] = set(vv for v in extra_deps.values() for vv in v)

    if key == "keys":
        return set(extra_deps.keys())
    else:
        return extra_deps[key]


def _ensure_list(value):
    if not value:
        return []
    elif isinstance(value, str):
        return [value]
    return list(value)


class eKonf:
    """ekorpkit config primary class"""

    __version__ = __version__()
    __ekorpkit_path__ = __ekorpkit_path__()
    __home_path__ = __home_path__()
    config = _config
    Keys = _Keys

    def __init__(self) -> None:
        raise NotImplementedError("Use one of the static construction functions")

    @staticmethod
    def compose(
        overrides: List[str] = [],
        config_group: str = None,
        *,
        return_as_dict: bool = False,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
        config_name="ekonf",
        verbose: bool = False,
    ):
        return _compose(
            overrides,
            config_group=config_group,
            return_as_dict=return_as_dict,
            throw_on_resolution_failure=throw_on_resolution_failure,
            throw_on_missing=throw_on_missing,
            config_name=config_name,
            verbose=verbose,
        )

    @staticmethod
    def select(
        cfg: Any,
        key: str,
        *,
        default: Any = None,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
    ):
        return _select(
            cfg,
            key,
            default=default,
            throw_on_resolution_failure=throw_on_resolution_failure,
            throw_on_missing=throw_on_missing,
        )

    @staticmethod
    def to_dict(
        cfg: Any,
    ):
        return _to_dict(cfg)

    @staticmethod
    def to_config(
        cfg: Any,
    ):
        return _to_config(cfg)

    @staticmethod
    def to_yaml(cfg: Any, *, resolve: bool = False, sort_keys: bool = False) -> str:
        if resolve:
            cfg = _to_dict(cfg)
        return _to_yaml(cfg, resolve=resolve, sort_keys=sort_keys)

    @staticmethod
    def to_container(
        cfg: Any,
        *,
        resolve: bool = False,
        throw_on_missing: bool = False,
        enum_to_str: bool = False,
        structured_config_mode: SCMode = SCMode.DICT,
    ):
        return _to_container(
            cfg,
            resolve,
            throw_on_missing,
            enum_to_str,
            structured_config_mode,
        )

    @staticmethod
    def partial(
        config: Any = None, config_group: str = None, *args: Any, **kwargs: Any
    ) -> Any:
        return _partial(config=config, config_group=config_group, *args, **kwargs)

    @staticmethod
    def instantiate(config: Any, *args: Any, **kwargs: Any) -> Any:
        return _instantiate(config, *args, **kwargs)

    @staticmethod
    def is_config(
        cfg: Any,
    ):
        return _is_config(cfg)

    @staticmethod
    def is_instantiatable(cfg: Any):
        return _is_instantiatable(cfg)

    @staticmethod
    def load(file_: Union[str, pathlib.Path, IO[Any]]) -> Union[DictConfig, ListConfig]:
        return _load(file_)

    @staticmethod
    def update(_dict, _overrides):
        return _update(_dict, _overrides)

    @staticmethod
    def merge(
        *configs: Union[
            DictConfig,
            ListConfig,
            Dict[DictKeyType, Any],
            List[Any],
            Tuple[Any, ...],
            Any,
        ],
    ) -> Union[ListConfig, DictConfig]:
        """
        Merge a list of previously created configs into a single one
        :param configs: Input configs
        :return: the merged config object.
        """
        return _merge(*configs)

    @staticmethod
    def save(
        config: Any, f: Union[str, pathlib.Path, IO[Any]], resolve: bool = False
    ) -> None:
        _save(config, f, resolve)

    @staticmethod
    def pprint(cfg: Any, resolve: bool = True, **kwargs):
        _print(cfg, resolve=resolve, **kwargs)

    @staticmethod
    def print(cfg: Any, resolve: bool = True, **kwargs):
        _print(cfg, resolve=resolve, **kwargs)

    @staticmethod
    def methods(cfg: Any, obj: object):
        _methods(cfg, obj)

    @staticmethod
    def function(cfg: Any, _name_, return_function=False, **parms):
        return _function(cfg, _name_, return_function, **parms)

    @staticmethod
    def run(config: Any, **kwargs: Any) -> Any:
        _run(config, **kwargs)

    @staticmethod
    def _load_dotenv(verbose: bool = False):
        _load_dotenv(verbose)

    @staticmethod
    def _init_env_(cfg, verbose=False):
        _init_env_(cfg, verbose=verbose)

    @staticmethod
    def _stop_env_(cfg, verbose=False):
        _stop_env_(cfg, verbose=verbose)

    @staticmethod
    def path(
        url_or_filename,
        extract_archive: bool = False,
        force_extract: bool = False,
        cache_dir=None,
    ):
        """
        Given something that might be a URL or local path, determine which.
        If it's a remote resource, download the file and cache it, and
        then return the path to the cached file. If it's already a local path,
        make sure the file exists and return the path.

        For URLs, the following schemes are all supported out-of-the-box:

        * ``http`` and ``https``,
        * ``s3`` for objects on `AWS S3`_,
        * ``gs`` for objects on `Google Cloud Storage (GCS)`_, and
        * ``gd`` for objects on `Google Drive`_, and
        * ``hf`` for objects or repositories on `HuggingFace Hub`_.

        Examples
        --------

        To download a file over ``https``::

            cached_path("https://github.com/allenai/cached_path/blob/main/README.md")

        To download an object on GCS::

            cached_path("gs://allennlp-public-models/lerc-2020-11-18.tar.gz")

        To download the PyTorch weights for the model `epwalsh/bert-xsmall-dummy`_
        on HuggingFace, you could do::

            cached_path("hf://epwalsh/bert-xsmall-dummy/pytorch_model.bin")

        For paths or URLs that point to a tarfile or zipfile, you can append the path
        to a specific file within the archive to the ``url_or_filename``, preceeded by a "!".
        The archive will be automatically extracted (provided you set ``extract_archive`` to ``True``),
        returning the local path to the specific file. For example::

            cached_path("model.tar.gz!weights.th", extract_archive=True)

        .. _epwalsh/bert-xsmall-dummy: https://huggingface.co/epwalsh/bert-xsmall-dummy

        Parameters
        ----------

        url_or_filename :
            A URL or path to parse and possibly download.

        extract_archive :
            If ``True``, then zip or tar.gz archives will be automatically extracted.
            In which case the directory is returned.

        force_extract :
            If ``True`` and the file is an archive file, it will be extracted regardless
            of whether or not the extracted directory already exists.

            .. caution::
                Use this flag with caution! This can lead to race conditions if used
                from multiple processes on the same file.

        cache_dir :
            The directory to cache downloads. If not specified, the global default cache directory
            will be used (``~/.cache/cached_path``). This can be set to something else with
            :func:`set_cache_dir()`.

        Returns
        -------
        :class:`pathlib.Path`
            The local path to the (potentially cached) resource.

        Raises
        ------
        ``FileNotFoundError``

            If the resource cannot be found locally or remotely.

        ``ValueError``
            When the URL is invalid.

        ``Other errors``
            Other error types are possible as well depending on the client used to fetch
            the resource.

        """
        return _path(
            url_or_filename,
            extract_archive=extract_archive,
            force_extract=force_extract,
            cache_dir=cache_dir,
        )

    @staticmethod
    def pipe(cfg, data=None):
        return apply_pipe(data, cfg)

    @staticmethod
    def dependencies(key, path=None):
        return _dependencies(key, path)

    @staticmethod
    def ensure_list(value):
        return _ensure_list(value)

    @staticmethod
    def to_dateparm(_date, _format="%Y-%m-%d"):
        return _to_dateparm(_date, _format)



import os
import pathlib
import logging

log = logging.getLogger(__name__)


def cached_path(
    url_or_filename,
    extract_archive: bool = False,
    force_extract: bool = False,
    return_dir: bool = False,
    cache_dir=None,
    verbose: bool = False,
):
    import cached_path

    if verbose:
        log.info(
            "caching path: {}, extract_archive: {}, force_extract: {}, cache_dir: {}".format(
                url_or_filename, extract_archive, force_extract, cache_dir
            )
        )

    if url_or_filename:
        try:
            if url_or_filename.startswith("gd://"):

                _path = cached_gdown(
                    url_or_filename,
                    verbose=verbose,
                    extract_archive=extract_archive,
                    force_extract=force_extract,
                    cache_dir=cache_dir,
                )
            else:
                if cache_dir is None:
                    cache_dir = (
                        pathlib.Path.home() / ".ekorpkit" / ".cache" / "cached_path"
                    )
                else:
                    cache_dir = pathlib.Path(cache_dir) / "cached_path"

                _path = cached_path.cached_path(
                    url_or_filename,
                    extract_archive=extract_archive,
                    force_extract=force_extract,
                    cache_dir=cache_dir,
                ).as_posix()

            log.info(f"cached path: {_path}")
            if return_dir and pathlib.Path(_path).is_file():
                _path = pathlib.Path(_path).parent

            return _path

        except Exception as e:
            log.error(e)
            return None


def cached_gdown(
    url, verbose=False, extract_archive=None, force_extract=False, cache_dir=None
):
    """
    :type url: str
          ex) gd://id:path
    :type verbose: bool
    :type extract_archive: bool
    :type force_extract: bool
    :type cache_dir: str
    :returns: str
    """
    import gdown

    if verbose:
        log.info(f"Downloading {url}...")
    if cache_dir is None:
        cache_dir = pathlib.Path.home() / ".ekorpkit" / ".cache" / "gdown"
    else:
        cache_dir = pathlib.Path(cache_dir) / "gdown"
    cache_dir.mkdir(parents=True, exist_ok=True)

    gd_prefix = "gd://"
    if url.startswith(gd_prefix):
        url = url[len(gd_prefix) :]
        _url = url.split(":")
        if len(_url) == 2:
            id, path = _url
        else:
            id = _url[0]
            path = id

        # If we're using the path!c/d/file.txt syntax, handle it here.
        fname = None
        extraction_path = path
        exclamation_index = path.find("!")
        if extract_archive and exclamation_index >= 0:
            extraction_path = path[:exclamation_index]
            fname = path[exclamation_index + 1 :]

        cache_path = cache_dir / f".{id}" / extraction_path
        cache_path.parent.mkdir(parents=True, exist_ok=True)

        cache_path = gdown.cached_download(
            id=id,
            path=cache_path.as_posix(),
            quiet=not verbose,
        )

        if extract_archive:
            extraction_path, files = extractall(cache_path, force_extract=force_extract)

            if fname and files:
                for f in files:
                    if f.endswith(fname):
                        return f
            else:
                return extraction_path
        return cache_path

    else:
        log.warning(f"Unknown url: {url}")
        return None


def extractall(path, to=None, force_extract=False):
    """Extract archive file.

    Parameters
    ----------
    path: str
        Path of archive file to be extracted.
    to: str, optional
        Directory to which the archive file will be extracted.
        If None, it will be set to the parent directory of the archive file.
    """
    import tarfile
    from zipfile import ZipFile

    if to is None:
        to = os.path.dirname(path)

    if path.endswith(".zip"):
        opener, mode = ZipFile, "r"
    elif path.endswith(".tar"):
        opener, mode = tarfile.open, "r"
    elif path.endswith(".tar.gz") or path.endswith(".tgz"):
        opener, mode = tarfile.open, "r:gz"
    elif path.endswith(".tar.bz2") or path.endswith(".tbz"):
        opener, mode = tarfile.open, "r:bz2"
    else:
        log.warning(
            "Could not extract '%s' as no appropriate " "extractor is found" % path
        )
        return path, None

    def namelist(f):
        if isinstance(f, ZipFile):
            return f.namelist()
        return [m.path for m in f.members]

    def filelist(f):
        files = []
        for fname in namelist(f):
            fname = os.path.join(to, fname)
            files.append(fname)
        return files

    extraction_name = pathlib.Path(path).stem
    extraction_path = f"{to}/{extraction_name}"
    if extraction_path is not None:
        # If the extracted directory already exists (and is non-empty), then no
        # need to extract again unless `force_extract=True`.
        if (
            os.path.isdir(extraction_path)
            and os.listdir(extraction_path)
            and not force_extract
        ):
            files = [
                os.path.join(dirpath, filename)
                for dirpath, _, filenames in os.walk(extraction_path)
                for filename in filenames
            ]

            return to, files

    with opener(path, mode) as f:
        f.extractall(path=to)

    return to, filelist(f)



import logging
import codecs
import os
import numpy as np
import pandas as pd
import ekorpkit.utils.batch.batcher as batcher
from collections import OrderedDict
from functools import partial, reduce
from ekorpkit.utils import print_status
from ekorpkit.utils.batch import decorator_apply
from ekorpkit.utils.func import check_max_len, check_min_len, elapsed_timer
from ekorpkit.io.file import save_dataframe as save_dataframe_
from ekorpkit import eKonf
from ekorpkit.ekonf import apply_pipe
from tqdm.auto import tqdm


log = logging.getLogger(__name__)


def apply(
    func,
    series,
    description=None,
    verbose=False,
    use_batcher=True,
    minibatch_size=None,
    num_workers=None,
    **kwargs,
):
    batcher_instance = batcher.batcher_instance
    if use_batcher and batcher_instance is not None:
        if batcher_instance is not None:
            batcher_minibatch_size = batcher_instance.minibatch_size
        else:
            batcher_minibatch_size = 1000
        if minibatch_size is None:
            minibatch_size = batcher_minibatch_size
        if num_workers is not None:
            batcher_instance.procs = num_workers
        if batcher_instance.procs > 1:
            batcher_instance.minibatch_size = min(
                int(len(series) / batcher_instance.procs) + 1, minibatch_size
            )
            log.info(
                f"Using batcher with minibatch size: {batcher_instance.minibatch_size}"
            )
            results = decorator_apply(func, batcher_instance, description=description)(
                series
            )
            if batcher_instance is not None:
                batcher_instance.minibatch_size = batcher_minibatch_size
            return results

    if batcher_instance is None:
        log.warning("Warning: batcher not initialized")
    tqdm.pandas(desc=description)
    return series.progress_apply(func)


def apply_pipeline(df, pipeline, pipeline_args, update_args={}, verbose=True):
    pipeline = eKonf.to_dict(pipeline)
    pipeline_args = eKonf.to_dict(pipeline_args)

    pipeline_targets = []
    pipes = OrderedDict()
    if isinstance(pipeline, list):
        for pipe in pipeline:
            pipes[pipe] = pipe
    elif isinstance(pipeline, str):
        pipes[pipeline] = pipeline
    else:
        pipes = pipeline

    if pipes is None or len(pipes) == 0:
        log.warning("No pipeline specified")
        return df

    log.info(f"Applying pipeline: {pipes}")
    for pipe, pipe_arg_name in pipes.items():
        args = pipeline_args.get(pipe_arg_name, {}).copy()
        if pipe != pipe_arg_name:
            args_override = pipeline_args.get(pipe, {})
            args.update(args_override)
        if args and isinstance(args, dict):
            args.update(update_args)
            pipeline_targets.append(args)

    return reduce(apply_pipe, pipeline_targets, df)


def split_column(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    source = args.get("source")
    target = args.get("target")
    _split = args.get(eKonf.Keys.SPLIT)
    if source is None:
        log.warning("No source specified")
        return df
    log.info(f"Split column: {args}")
    df[target] = df[source].str.split(**_split)

    if verbose:
        print(df.tail())
    return df


def eval_columns(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    expressions = args.get("expressions", None)
    engine = args.get("engine", None)
    eval_at = args.get("eval_at", "dataframe")
    if expressions is None:
        log.warning("No expressions specified")
        return df
    log.info(f"Eval columns: {args}")
    if eval_at == "dataframe":
        if isinstance(expressions, list):
            for expr in expressions:
                df.eval(expr, engine=engine, inplace=True)
        else:
            for col, expr in expressions.items():
                df[col] = df.eval(expr, engine=engine)
    elif eval_at == "python":
        if isinstance(expressions, dict):
            for col, expr in expressions.items():
                df[col] = eval(expr)
        else:
            log.warning("Expressions must be a dict to eval at python")
    else:
        if isinstance(expressions, list):
            for expr in expressions:
                pd.eval(expr, engine=engine, inplace=True, target=df)
        else:
            for col, expr in expressions.items():
                df[col] = pd.eval(expr, engine=engine)

    if verbose:
        print(df.tail())
    return df


def combine_columns(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    columns = args.get("columns", None)
    if columns is None:
        log.warning("No columns specified")
        return df
    log.info(f"Combining columns: {args}")
    for col in columns:
        df[col].fillna("", inplace=True)
        df[col] = df[col].astype(str)
    separator = codecs.decode(args["separator"], "unicode_escape")
    df[args["into"]] = df[columns].agg(separator.join, axis=1)
    return df


def drop(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    labels = args.get("labels", None)
    axis = args.get("axis", 1)
    columns = args.get("columns", None)
    index = args.get("index", None)
    level = args.get("level", None)
    errors = args.get("errors", "ignore")
    if verbose:
        log.info(f"droping: {args}")
    df = df.drop(
        columns=columns,
        axis=axis,
        labels=labels,
        index=index,
        level=level,
        errors=errors,
    )
    if verbose:
        print(df.head())
    return df


def melt(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    id_vars = args.get("id_vars", None)
    value_vars = args.get("value_vars", None)
    var_name = args.get("var_name", None)
    value_name = args.get("value_name", None)
    col_level = args.get("col_level", None)
    ignore_index = args.get("ignore_index", True)
    if id_vars is None:
        if verbose:
            log.warning("No id_vars specified")
        return df
    if verbose:
        log.info(f"Melting columns: {args}")
    id_vars = list(id_vars)
    if value_vars:
        if isinstance(value_vars, str):
            value_vars = eval(value_vars)
        else:
            value_vars = list(value_vars)
    df = df.melt(
        id_vars=id_vars,
        value_vars=value_vars,
        var_name=var_name,
        value_name=value_name,
        col_level=col_level,
        ignore_index=ignore_index,
    )
    if verbose:
        print(df.head())
    return df


def plot(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    plot_cfg = args.get("visualize", {}).get("plot", {})
    if "_target_" not in plot_cfg:
        log.warning("No target specified")
        return df
    if verbose:
        log.info(f"Plotting: {plot_cfg}")
    subset = args.get("subset", {})
    col = subset.get("column", None)
    values = subset.get("values", None)
    titles = subset.get("titles", None)
    output_file = args.get("output_file", None)
    output_dir = args.get("output_dir", ".")
    if subset and col:
        title_prefix = plot_cfg["figure"].get("title", None)
        title_prefix = title_prefix + " - " if title_prefix else ""
        if isinstance(values, str):
            values = [values]
        if titles and isinstance(titles, str):
            titles = [titles]
        for i, val in enumerate(values):
            if verbose:
                log.info(f"Plotting subset: {col} == {val}")
            df_sub = df[df[col] == val]
            if output_file and "{}" in output_file:
                output_path = output_file.replace("{}", str(val))
                output_path = f"{output_dir}/{output_path}"
                plot_cfg["savefig"]["fname"] = output_path
            if titles and len(titles) > i:
                plot_cfg["figure"]["title"] = title_prefix + titles[i]
            if verbose:
                log.info(f"Plotting: {plot_cfg}")
                # print(df_sub.head())
            eKonf.instantiate(plot_cfg, df=df_sub)
    else:
        eKonf.instantiate(plot_cfg, df=df)

    return df


def pivot(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    index = args.get("index", None)
    columns = args.get("columns", None)
    values = args.get("values", None)
    fillna = args.get("fillna", None)
    reset_index = args.get("reset_index", True)
    if index is None:
        log.warning("No index specified")
        return df
    if columns is None:
        log.warning("No columns specified")
        return df
    if values is None:
        log.warning("No values specified")
        return df
    if verbose:
        log.info(f"Pivoting columns: {args}")
    if len(values) == 1:
        values = values[0]
    df = df.pivot(index=index, columns=columns, values=values)
    # df = df.set_index(index)[values].unstack()
    if fillna is not None:
        df.fillna(fillna, inplace=True)
    if reset_index:
        if verbose:
            log.info(f"Resetting index, nlevels of columns: {df.columns.nlevels}")
        df.reset_index(inplace=True)
        if df.columns.nlevels > 1:
            df.columns = ["_".join(a).strip("_") for a in df.columns.to_flat_index()]
    if verbose:
        print(df.head())
    return df


def split_sampling(df, args):
    from sklearn.model_selection import train_test_split

    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    stratify_on = args.get("stratify_on", None)
    random_state = args.get("random_state", 123)
    groupby = args.get("groupby", stratify_on)
    unique_key = args.get("unique_key", "id")
    test_size = args.get("test_size", 0.1)
    dev_size = args.get("dev_size", None)
    if verbose:
        log.info(f"Split sampling: {args}")

    train, dev, test = None, None, None
    if stratify_on is None:
        train, test = train_test_split(
            df, test_size=test_size, random_state=random_state
        )
        if dev_size:
            train, dev = train_test_split(
                train, test_size=dev_size, random_state=random_state
            )
    else:
        train, test = train_test_split(
            df, test_size=test_size, random_state=random_state, stratify=df[stratify_on]
        )
        if dev_size:
            train, dev = train_test_split(
                train,
                test_size=dev_size,
                random_state=random_state,
                stratify=train[stratify_on],
            )
    train.reset_index(drop=True, inplace=True)
    test.reset_index(drop=True, inplace=True)
    if dev_size:
        dev.reset_index(drop=True, inplace=True)
    if verbose:
        log.info(f"Total rows: {len(df)}")
        log.info(f"Train: {len(train)}")
        log.info(f"Test: {len(test)}")
        if dev_size:
            log.info(f"Dev: {len(dev)}")

        grp_all = (
            df.groupby(groupby)[unique_key]
            .count()
            .rename("population")
            .transform(lambda x: x / x.sum() * 100)
        )
        grp_train = (
            train.groupby(groupby)[unique_key]
            .count()
            .rename("train")
            .transform(lambda x: x / x.sum() * 100)
        )
        grp_test = (
            test.groupby(groupby)[unique_key]
            .count()
            .rename("test")
            .transform(lambda x: x / x.sum() * 100)
        )
        grp_dists = pd.concat([grp_all, grp_train, grp_test], axis=1)
        if dev_size:
            grp_dev = (
                dev.groupby(groupby)[unique_key]
                .count()
                .rename("dev")
                .transform(lambda x: x / x.sum() * 100)
            )
            grp_dists = pd.concat([grp_dists, grp_dev], axis=1)
        print(grp_dists)

    output_dir = args.get("output_dir", ".")
    train_file = args.get("train_file", None)
    test_file = args.get("test_file", None)
    dev_file = args.get("dev_file", None)
    if train_file and train is not None:
        save_dataframe_(
            train, output_dir=output_dir, output_file=train_file, verbose=verbose
        )
    if test_file and test is not None:
        save_dataframe_(
            test, output_dir=output_dir, output_file=test_file, verbose=verbose
        )
    if dev_file and dev is not None:
        save_dataframe_(
            dev, output_dir=output_dir, output_file=dev_file, verbose=verbose
        )

    return df


def top_values(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    groupby = args.get("groupby", None)
    value_var = args.get("value_var", None)
    value_label = args.get("value_label", None)
    value_name = args.get("value_name", None)
    value_separator = args.get("value_separator", ", ")
    top_n = args.get("top_n", 5)
    columns = args.get("columns", None)
    use_batcher = args.get("use_batcher", True)
    minibatch_size = args.get("minibatch_size", None)

    if verbose:
        log.info(f"Split sampling: {args}")

    def label(row):
        return f"{row[value_label]}[{round(row[value_var]*100,0):.0f}%]"

    if value_label is not None and value_name is not None:
        # df[value_name] = df.apply(label, axis=1)
        df[value_name] = apply(
            label,
            df,
            description="labeling",
            verbose=verbose,
            use_batcher=use_batcher,
            minibatch_size=minibatch_size,
        )

    top_n_grp = (
        df.sort_values(by=value_var, ascending=False)
        .groupby(groupby)
        .head(top_n)
        .reset_index(drop=True)
    )
    top_n_grp = top_n_grp.sort_values(by=groupby)
    top_n_grp = top_n_grp.groupby(groupby, as_index=False).agg(
        {value_name: value_separator.join}
    )
    top_n_grp = top_n_grp[groupby + [value_name]]
    df.drop(columns=value_name, inplace=True)
    df = pd.merge(df, top_n_grp, on=groupby, how="left")
    if columns:
        df = df[columns:]
    if verbose:
        print(df.head())

    return df


def subset(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    random_state = args.get("random_state", 123)
    head_n = args.get("head_n", None)
    tail_n = args.get("tail_n", None)
    sample_n = args.get("sample_n", None)
    sample_frac = args.get("sample_frac", None)
    if not (head_n or tail_n or sample_n or sample_frac):
        log.error("Must specify one of head_n, tail_n, sample_n, sample_frac")
        return df

    if verbose:
        log.info(f"Subsetting: {args}")
    dfs = []
    if head_n:
        dfs.append(df.head(head_n))
    if tail_n:
        dfs.append(df.tail(tail_n))
    if sample_n:
        dfs.append(df.sample(sample_n, random_state=random_state))
    if sample_frac:
        dfs.append(df.sample(frac=sample_frac, random_state=random_state))
    if verbose:
        log.info(f"Total rows: {len(df)}")

    if len(dfs) > 0:
        df = pd.concat(dfs)

    if verbose:
        log.info(f"Subset rows: {len(df)}")

    return df


def sampling(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    random_state = args.get("random_state", 123)
    groupby = args.get("groupby", None)
    sample_size_per_group = args.get("sample_size_per_group", None)
    value_var = args.get("value_var", None)
    columns = args.get("columns", None)

    if verbose:
        log.info(f"Split sampling: {args}")

    if groupby is None:
        if sample_size_per_group < 1:
            df_sample = df.sample(frac=sample_size_per_group, random_state=random_state)
        else:
            df_sample = df.sample(n=sample_size_per_group, random_state=random_state)
    else:
        if sample_size_per_group < 1:
            df_sample = df.groupby(groupby, group_keys=False).apply(
                lambda x: x.sample(
                    frac=sample_size_per_group, random_state=random_state
                )
            )
        else:
            df_sample = df.groupby(groupby, group_keys=False).apply(
                lambda x: x.sample(
                    n=min(len(x), sample_size_per_group), random_state=random_state
                )
            )
    df_sample.reset_index(drop=True, inplace=True)
    if columns:
        df_sample = df_sample[columns:]
    if verbose:
        log.info(f"Total rows: {len(df)}")
        log.info(f"Sample: {len(df_sample)}")
        print(df_sample.head())

    if groupby is not None and verbose:
        grp_all = df.groupby(groupby)[value_var].count().rename("population")
        grp_sample = df_sample.groupby(groupby)[value_var].count().rename("sample")
        grp_dists = pd.concat([grp_all, grp_sample], axis=1)
        print(grp_dists)

    save_dataframe(df_sample, args)

    return df


def aggregate_columns(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    onto_column = args.get("onto", None)
    aggregations = args.get("aggregations", None)
    reset_index = args.get("reset_index", False)
    if onto_column is None and aggregations is None:
        if verbose:
            log.warning("No columns or aggregations are specified")
        return df
    groupby_cloumns = args["groupby"]
    if groupby_cloumns is None:
        if verbose:
            log.warning("No groupby specified")
        return df
    separator = codecs.decode(args["separator"], "unicode_escape")

    num_docs = df.shape[0]
    if verbose:
        log.info(f"Aggregating columns: {args}")
    if aggregations:
        df = df.groupby(groupby_cloumns, as_index=False).agg(aggregations)
    else:
        df[onto_column].fillna("", inplace=True)
        df = df.groupby(groupby_cloumns, as_index=False).agg(
            {onto_column: separator.join}
        )
    if reset_index:
        if verbose:
            log.info(f"Resetting index, nlevels of columns: {df.columns.nlevels}")
        df.reset_index(inplace=True)
        if df.columns.nlevels > 1:
            df.columns = ["_".join(a).strip("_") for a in df.columns.to_flat_index()]
    n_docs = df.shape[0]
    if verbose:
        print(df.tail())
        log.info(f"{num_docs} documents aggregated into {n_docs} documents")
    return df


def explode_splits(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    separator = codecs.decode(args["separator"], "unicode_escape")
    id_key = args.get("id_key", "id")
    split_key = args.get("split_key", "seg_id")
    if verbose:
        log.info(f"Exploding column: {args}")

    num_docs = df.shape[0]
    df[apply_to] = df[apply_to].str.split(separator)
    df = df.explode(apply_to)
    df[split_key] = df.groupby(id_key).cumcount()
    n_docs = df.shape[0]
    if verbose:
        log.info(f"{num_docs} documents exploded into {n_docs} documents")
    return df


def rename_columns(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    new_names = args.get("new_names", None)
    if new_names is None:
        if verbose:
            log.warning("No columns specified")
        return df
    if verbose:
        log.info(f"Renaming columns: {args}")
    if new_names is not None:
        df.rename(columns=new_names, inplace=True)
    if verbose:
        print(df.head())
    return df


def reset_index(df, args):
    verbose = args.get("verbose", False)
    index_column_name = args.get("index_column_name", None)
    if index_column_name is None:
        index_column_name = "index"
    drop_index = args.get("drop_index", False)
    if verbose:
        log.info(f"Resetting index: {args}")
    df = df.reset_index(drop=drop_index)
    if not drop_index and index_column_name != "index":
        df.rename(columns={"index": index_column_name}, inplace=True)
    if verbose:
        print(df.head())
    return df


def normalize(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        log.warning("No columns specified")
        return df
    normalizer = args.get("preprocessor", {}).get("normalizer", None)
    if normalizer is None:
        log.warning("No normalizer specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    log.info("instantiating normalizer")
    normalizer = eKonf.instantiate(normalizer)
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            df[key] = apply(
                normalizer.normalize,
                df[key],
                description=f"Normalizing column: {key}",
                verbose=verbose,
            )
            log.info(" >> elapsed time to normalize: {}".format(elapsed()))
    return df


def fillna(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    fill_with = args.get("fill_with", "")
    if verbose:
        log.info(f"Filling missing values: {args}")
    for key in apply_to:
        df[key].fillna(fill_with, inplace=True)
    return df


def predict(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    num_workers = args.get("num_workers", 1)
    use_batcher = args.get("use_batcher", True)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    columns = args["columns"]

    to_predict = args.get("to_predict", None)
    model = args.get("model", None)
    if model is None:
        log.warning("No model specified")
        return df

    if isinstance(apply_to, list):
        apply_to = apply_to[0]

    log.info("instantiating model")

    _target_ = model[eKonf.Keys.TARGET]
    model = eKonf.instantiate(model)

    if "sentiment.analyser" in _target_:
        key = apply_to
        with elapsed_timer(format_time=True) as elapsed:
            predictions = apply(
                model.predict,
                df[key],
                description=f"Predicting [{key}]",
                verbose=verbose,
                use_batcher=use_batcher,
                num_workers=num_workers,
            )
            pred_df = pd.DataFrame(predictions.tolist(), index=predictions.index)
            if columns:
                columns += pred_df.columns.tolist()
                args["columns"] = columns
            df = df.join(pred_df)
            log.info(" >> elapsed time to predict: {}".format(elapsed()))
    else:
        df = model.predict(df, to_predict)
        if columns:
            columns.append(model._to_predict["predicted"])
            args["columns"] = columns

    save_dataframe(df, args)

    return df


def segment(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    use_batcher = args.get("use_batcher", True)
    minibatch_size = args.get("minibatch_size", None)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    segmenter = args.get("preprocessor", {}).get("segmenter", None)
    if segmenter is None:
        log.warning("No segmenter specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    log.info("instantiating segmenter")
    segmenter = eKonf.instantiate(segmenter)
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            df[key] = apply(
                segmenter.segment_article,
                df[key],
                description=f"Splitting column: {key}",
                verbose=verbose,
                use_batcher=use_batcher,
                minibatch_size=minibatch_size,
            )
            log.info(" >> elapsed time to segment: {}".format(elapsed()))
    return df


def tokenize(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    use_batcher = args.get("use_batcher", True)
    minibatch_size = args.get("minibatch_size", None)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        log.warning("No columns specified")
        return df
    tokenizer = args.get("preprocessor", {}).get("tokenizer", None)
    if tokenizer is None:
        log.warning("No tokenizer specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    log.info("instantiating tokenizer")
    tokenizer = eKonf.instantiate(tokenizer)
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            df[key] = apply(
                tokenizer.tokenize_article,
                df[key],
                description=f"Tokenizing column: {key}",
                verbose=verbose,
                use_batcher=use_batcher,
                minibatch_size=minibatch_size,
            )
            log.info(" >> elapsed time to segment: {}".format(elapsed()))
    return df


def extract_tokens(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    use_batcher = args.get("use_batcher", True)
    nouns_only = args.get("nouns_only", True)
    filter_stopwords_only = args.get("filter_stopwords_only", False)
    minibatch_size = args.get("minibatch_size", None)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        log.warning("No columns specified")
        return df
    tokenizer = args.get("preprocessor", {}).get("tokenizer", None)
    if tokenizer is None:
        log.warning("No tokenizer specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    log.info("instantiating tokenizer")
    tokenizer = eKonf.instantiate(tokenizer)
    if filter_stopwords_only:
        extract_func = tokenizer.filter_article_stopwords
    elif nouns_only:
        extract_func = tokenizer.extract_nouns
    else:
        extract_func = tokenizer.extract_tokens
    log.info(f"extract_func: {extract_func}")

    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            df[key] = apply(
                extract_func,
                df[key],
                description=f"Extracting column: {key}",
                verbose=verbose,
                use_batcher=use_batcher,
                minibatch_size=minibatch_size,
            )
            log.info(" >> elapsed time to extract tokens: {}".format(elapsed()))
    return df


def chunk(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    use_batcher = args.get("use_batcher", True)
    minibatch_size = args.get("minibatch_size", None)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        log.warning("No columns specified")
        return df
    segmenter = args.get("preprocessor", {}).get("segmenter", None)
    if segmenter is None:
        log.warning("No segmenter specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    log.info("instantiating segmenter")
    segmenter = eKonf.instantiate(segmenter)
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            df[key] = apply(
                segmenter.chunk_article,
                df[key],
                description=f"Chunking column: {key}",
                verbose=verbose,
                use_batcher=use_batcher,
                minibatch_size=minibatch_size,
            )
            log.info(" >> elapsed time to segment: {}".format(elapsed()))
    return df


def general_function(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", None)
    method = args.get(eKonf.Keys.METHOD, None)
    if verbose:
        log.info(f"Running dataframe function: {args}")

    with elapsed_timer(format_time=True) as elapsed:
        if apply_to is None:
            df = getattr(df, method[eKonf.Keys.NAME])(**method[eKonf.Keys.PARMS])
        else:
            if isinstance(apply_to, str):
                apply_to = [apply_to]
            for key in apply_to:
                log.info(f"processing column: {key}")
                df[key] = getattr(df[key], method[eKonf.Keys.NAME])(
                    **method[eKonf.Keys.PARMS]
                )

        log.info(" >> elapsed time to replace: {}".format(elapsed()))
        if verbose:
            print(df.head())
    return df


def replace_regex(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    patterns = args.get("patterns", {})
    if patterns is None:
        if verbose:
            log.warning("No patterns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    if verbose:
        log.info(f"Replacing regex: {args}")
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            for pat, repl in patterns.items():
                df[key] = df[key].str.replace(pat, repl, regex=True).str.strip()
            if verbose:
                log.info(" >> elapsed time to replace regex: {}".format(elapsed()))
    return df


def remove_startswith(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    startswith = args.get("startswith", {})
    if startswith is None:
        if verbose:
            log.warning("No startswith text specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    if verbose:
        log.info(f"Remove startswith: {args}")
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            for starting_text in startswith:
                log.info(f"Remove text starting with {starting_text} from [{key}]")
                idx = df[key].str.lower().str.startswith(starting_text, na=False)
                start_pos = len(starting_text)
                df.loc[idx, key] = df.loc[idx, key].str[start_pos:].str.strip()
            if verbose:
                log.info(
                    " >> elapsed time to remove startswith: {}\n".format(elapsed())
                )
    return df


def filter_length(df, args, **kwargs):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    min_length = args.get("min_length", None)
    max_length = args.get("max_length", None)
    if min_length is None and max_length is None:
        if verbose:
            log.warning("No length specified")
        return df
    len_func = args[eKonf.Keys.FUNC].get("len_bytes", None)
    len_func = eKonf.instantiate(len_func)
    _check_max_len = partial(check_max_len, max_len=max_length, len_func=len_func)
    _check_min_len = partial(check_min_len, min_len=min_length, len_func=len_func)

    if verbose:
        log.info(f"Filtering by length: {args}")
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            if min_length and min_length > 0:
                n_docs = df.shape[0]
                idx = apply(
                    _check_min_len,
                    df[key].astype(str),
                    verbose=verbose,
                    description=f"min length: {min_length}",
                )
                df = df[idx]
                if verbose:
                    log.info(
                        f"{(n_docs-df.shape[0])} of {n_docs} documents removed due to length is less than {min_length}"
                    )
            if max_length and max_length > 0:
                n_docs = df.shape[0]
                idx = apply(
                    _check_max_len,
                    df[key].astype(str),
                    verbose=verbose,
                    description=f"max length: {max_length}",
                )
                df = df[idx]
                if verbose:
                    log.info(
                        f"{(n_docs-df.shape[0])} of {n_docs} documents removed due to length is greater than {max_length}"
                    )
            if verbose:
                log.info(" >> elapsed time to filter length: {}".format(elapsed()))
    return df


def filter_query(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    query = args.get("query", None)
    if query is None:
        if verbose:
            log.warning("No query specified")
        return df
    if isinstance(query, str):
        query = [query]

    if verbose:
        log.info(f"Filtering by qeury: {args}")
    with elapsed_timer(format_time=True) as elapsed:
        for qry in query:
            if verbose:
                log.info(f"Preprocessing query: {qry}")
            n_docs = df.shape[0]
            df = df.query(qry, engine="python")
            if verbose:
                log.info(f"filtered {df.shape[0]} out of {n_docs} documents by {qry}")
        if verbose:
            log.info(" >> elapsed time to filter query: {}".format(elapsed()))
    return df


def drop_duplicates(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", None)
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    if verbose:
        log.info(f"Dropping duplicates: {args}")
    with elapsed_timer(format_time=True) as elapsed:
        num_docs = df.shape[0]
        df = df.drop_duplicates(subset=apply_to)
        n_docs = df.shape[0]
        if verbose:
            log.info(
                f"{n_docs} documents after dropping {(num_docs-n_docs)} duplicates from [{apply_to}]"
            )
            log.info(" >> elapsed time to drop duplicates: {}".format(elapsed()))
    return df


def save_samples(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    sample_length_to_print = args.get("sample_length_to_print", 1000)
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    num_samples_to_save = args.get("num_samples_to_save", None)
    smaple_file_prefix = args.get("sample_file_prefix", "sample")
    if verbose:
        log.info(f"Saving samples: {args}")

    sample_separator = "-" * 100 + "\n"
    df_sample = df.sample(num_samples_to_save)[apply_to]
    sample_text = ""
    print_text = ""
    for i, row in df_sample.iterrows():
        for key in apply_to:
            stext = row[key]
            if len(stext) > sample_length_to_print:
                ptext = stext[:sample_length_to_print] + "..."
            else:
                ptext = stext
            sample_text += key + ": \n" + stext + "\n\n"
            print_text += key + ": \n" + ptext + "\n\n"
        sample_text += sample_separator
        print_text += sample_separator
    sample_text = sample_text.strip()
    print_text = print_text.strip()
    sample_file = smaple_file_prefix + ".txt"
    open(sample_file, "w", encoding="utf-8").write(sample_text)

    if verbose:
        print(sample_separator)
        print(print_text)
        log.info(f"Saved {num_samples_to_save} samples to {sample_file}")

    return df


def stdout_samples(df, args):
    from contextlib import redirect_stdout

    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    sample_length_to_print = args.get("sample_length_to_print", 1000)
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    num_samples = args.get("num_samples", None)
    output_dir = args.get("output_dir", None)
    if output_dir is None:
        output_dir = "."
    output_file = args.get("output_file", None)
    head = args.get("head", None)
    tail = args.get("tail", None)
    if head is None:
        head = 5
    if tail is None:
        tail = 5

    if verbose:
        log.info(f"Print samples: {args}")

    sample_separator = "-" * 100 + "\n"
    df_sample = df.sample(num_samples)[apply_to]

    print_text = ""
    for i, row in df_sample.iterrows():
        for key in apply_to:
            ptext = row[key]
            print_text += key + ": \n" + ptext + "\n\n"
        print_text += sample_separator
    print_text = print_text.strip()

    if output_file is not None:
        output_file = os.path.join(output_dir, output_file)
        with open(output_file, "w", encoding="utf-8") as f:
            with redirect_stdout(f):
                print(print_text)
                print(sample_separator)
                print(df.head(head))
                print(sample_separator)
                print(df.tail(tail))
    print(print_text)
    print(sample_separator)
    print(df.head(head))
    print(sample_separator)
    print(df.tail(tail))

    if verbose:
        log.info(f"Saved {num_samples} samples to {output_file}")

    return df


def save_as_text(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    corpus_name = args.get("corpus_name", eKonf.Keys.CORPUS)
    output_dir = args.get("output_dir", ".")
    output_file = args.get("output_file", None)
    doc_separator = args.get("doc_separator", "\n\n")
    if isinstance(apply_to, list):
        apply_to = apply_to[0]

    os.makedirs(os.path.abspath(output_dir), exist_ok=True)
    if output_file:
        filename = output_file
    else:
        filename = f"{corpus_name}.txt"
    output_file_path = f"{output_dir}/{filename}"

    with elapsed_timer(format_time=True) as elapsed:
        n_loaded = len(df.index)
        doc_separator = str(doc_separator).encode("utf-8").decode("unicode_escape")
        with open(output_file_path, "w") as fo:
            fo.write(doc_separator.join(df[apply_to].dropna().tolist()))
        log.info(f"Corpus is exported to {output_file_path}")
        n_sampled = len(df.index)
        status = [[" x ", corpus_name, n_loaded, n_sampled, elapsed(), filename]]

    if verbose:
        print_status(status)
    return df


def split_dataframe(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    num_splits = args.get("num_splits", 1)
    if num_splits <= 1:
        return df
    if verbose:
        log.info(f"Splitting dataframe into {num_splits} splits")
    return np.array_split(df, num_splits)


def concat_dataframes(dataframes, args):
    from ekorpkit.io.file import concat_dataframes as concat_dataframes_

    args = eKonf.to_dict(args)

    return concat_dataframes_(data=dataframes, **args)


def merge_dataframe(df=None, args=None):
    if args is None:
        raise ValueError("args must be specified")
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    how = args.get("how", "inner")
    merge_on = args.get("merge_on", None)
    left_on = args.get("left_on", None)
    right_on = args.get("right_on", None)
    if merge_on is None and (right_on is None or left_on is None):
        raise ValueError("merge_on or (left_on and right_on) must be specified")
    if isinstance(merge_on, str):
        merge_on = [merge_on]
    if isinstance(left_on, str):
        left_on = [left_on]
    if isinstance(right_on, str):
        right_on = [right_on]

    df_to_merge = load_dataframe(None, args)
    if merge_on:
        df = df.merge(df_to_merge, how=how, on=merge_on)
    else:
        df = df.merge(df_to_merge, how=how, left_on=left_on, right_on=right_on)
    if verbose:
        print(df.tail())
    return df


def save_metadata(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    filepath = args.get("filepath", None)
    filetype = args.get("filetype", None)
    column_info = args.get("column_info", None)
    split_name = args.get("split_name", None)

    if verbose:
        log.info(f"Saving metadata: {args}")

    meta_info = column_info.get("meta", None)
    meta_columns = []
    if isinstance(meta_info, dict):
        meta_columns = list(meta_info.keys())
        if eKonf.Keys.SPLIT in meta_columns and eKonf.Keys.SPLIT not in df.columns:
            df[eKonf.Keys.SPLIT] = split_name
        df_meta = df[meta_columns]
        save_dataframe_(df_meta, filepath, filetype, verbose)

    data_info = column_info.get("data", None)
    if isinstance(data_info, dict):
        data_keys = list(data_info.keys())
        # if eKonf.Keys.SPLIT in data_columns and eKonf.Keys.SPLIT not in df.columns:
        #     df[eKonf.Keys.SPLIT] = split_name
        data_columns = [
            col for col in df.columns if col not in meta_columns or col in data_keys
        ]
        df = df[data_columns]

    return df


def save_dataframe(df, args):
    args = eKonf.to_dict(args)

    if df is None:
        log.warning("Dataframe is None")
        return df

    save_dataframe_(df, **args)
    return df


def load_dataframe(df=None, args=None):
    from ekorpkit.io.file import get_filepaths, load_dataframe as _load_dataframe

    if args is None:
        raise ValueError("args must be specified")

    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    concatenate = args.get("concatenate", True)
    filepath = args.get("filepath", None)
    data_dir = args.get("data_dir", None)
    data_file = args.get("data_file", None)
    dtype = args.get("dtype", None)
    if isinstance(dtype, list):
        dtype = {k: "str" for k in dtype}
    parse_dates = args.get("parse_dates", False)
    columns = args.get("columns", None)

    if filepath:
        filepaths = get_filepaths(filepath)
    else:
        filepaths = get_filepaths(data_file, data_dir)
    log.info(f"Loading {len(filepaths)} dataframes from {filepaths}")
    if len(filepaths) == 1:
        df = _load_dataframe(
            filepaths[0],
            verbose=verbose,
            dtype=dtype,
            parse_dates=parse_dates,
            columns=columns,
        )
        return df
    else:
        if concatenate:
            df = pd.concat(
                [
                    _load_dataframe(
                        f,
                        verbose=verbose,
                        dtype=dtype,
                        parse_dates=parse_dates,
                        columns=columns,
                    )
                    for f in filepaths
                ]
            )
            return df
        else:
            df_dict = {}

            for f in filepaths:
                df = _load_dataframe(
                    f,
                    verbose=verbose,
                    dtype=dtype,
                    parse_dates=parse_dates,
                    columns=columns,
                )
                df_name = os.path.basename(f)
                df_dict[df_name] = df
            return df_dict


def summary_stats(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    output_dir = args.get("output_dir", ".")
    output_file = args.get("output_file", None)
    stat_args = args.get("info", {}).get("stats", None)
    if stat_args is None:
        log.warning("No stats specified")
        return df
    if output_file is None:
        output_file = "summary_stats.csv"
    os.makedirs(os.path.abspath(output_dir), exist_ok=True)
    info_path = f"{output_dir}/{output_file}"
    stat_fn = eKonf.instantiate(stat_args)
    stats = stat_fn(df)
    eKonf.save(stats, f=info_path)
    log.info(f"Saving summary stats: {info_path}")
    if verbose:
        eKonf.pprint(stats)

    return df


def save_as_json(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    corpus_name = args.get("corpus_name", eKonf.Keys.CORPUS)
    output_dir = args.get("output_dir", ".")
    output_file = args.get("output_file", None)
    force_ascii = args.get("force_ascii", False)

    os.makedirs(os.path.abspath(output_dir), exist_ok=True)
    if output_file:
        filename = output_file
    else:
        filename = f"{corpus_name}.json"
    output_file_path = f"{output_dir}/{filename}"

    df.to_json(output_file_path, orient="records", lines=True, force_ascii=force_ascii)
    log.info(f"Corpus is exported to {output_file_path}")
    return df


def pipeline(data=None, **cfg):
    args = eKonf.to_dict(cfg)
    verbose = args.get("verbose", False)

    if eKonf.is_instantiatable(data):
        data = eKonf.instantiate(data)

    df = data.data
    if df is None:
        raise ValueError("No dataframe to process")

    process_pipeline = args.get(eKonf.Keys.PIPELINE, [])
    if process_pipeline is None:
        process_pipeline = []

    if len(process_pipeline) > 0:
        df = apply_pipeline(df, process_pipeline, args)
    else:
        log.warning("No pipeline specified")

    return df



import os
import re
import logging
from statistics import mean
from scipy.stats import pearsonr, spearmanr
from transformers.data.metrics.squad_metrics import compute_exact, compute_f1
from .simple import SimpleTrainer


log = logging.getLogger(__name__)


class SimpleT5(SimpleTrainer):
    def __init__(self, **args):
        super().__init__(**args)

    def load_datasets(self):
        super().load_datasets()
        cols = self.train_data.columns
        renames = {
            name: key for key, name in self._to_train.items() if name and name in cols
        }
        if renames:
            self.train_data.rename(columns=renames, inplace=True)
            if self.eval_data is not None:
                self.eval_data.rename(columns=renames, inplace=True)
            if self.test_data is not None:
                self.test_data.rename(columns=renames, inplace=True)
        task_prefix = self._to_train.get("task_prefix")
        if self._to_train["prefix"] is None:
            self.train_data["prefix"] = task_prefix
            if self.eval_data is not None:
                self.eval_data["prefix"] = task_prefix
            if self.test_data is not None:
                self.test_data["prefix"] = task_prefix
        if self.verbose:
            print("Train data for T5:")
            print(self.train_data.head())

    def train(self):
        from simpletransformers.t5 import T5Model

        args = self.args

        if not self.splits:
            self.load_datasets()

        # Create a Model
        model = T5Model(
            args["model_type"],
            args["model_uri"],
            cuda_device=args["cuda_device"],
            args=self._model_cfg,
        )

        # Train the model
        model.train_model(
            self.train_data,
            eval_data=self.eval_data,
        )

        # # Evaluate the model
        result = model.eval_model(self.test_data)
        print(result)

        # # Check predictions
        # # print(predictions[:5])
        # return result, model_outputs, predictions

    def load_model(self, model_dir=None, pred_args=None):
        from simpletransformers.t5 import T5Model

        if model_dir is None:
            model_dir = self.args["best_model_dir"]
        if pred_args is not None:
            self._model_cfg.update(pred_args)

        self.model = T5Model(self.args["model_type"], model_dir, args=self._model_cfg)
        log.info(f"Loaded model from {model_dir}")

    def _predict(self, to_predict: list):
        if self.model is None:
            self.load_model()

        preds = self.model.predict(to_predict)
        return preds

    def convert_to_predict(self, df):
        input_key = self._to_predict["input"]
        task_prefix = self._to_predict["task_prefix"]
        if self._to_predict["prefix"] is None:
            df["prefix"] = task_prefix

        to_predict = [
            prefix + ": " + str(input_text)
            for prefix, input_text in zip(df["prefix"].tolist(), df[input_key].tolist())
        ]

        if self.verbose:
            print(to_predict[:5])
        return to_predict

    def append_predictions(self, df, preds):
        predicted_key = self._to_predict["predicted"]
        if self._model_cfg["num_return_sequences"] > 1:
            preds = [pred[0] for pred in preds]
        df[predicted_key] = preds
        return df


def f1(truths, preds):
    return mean([compute_f1(truth, pred) for truth, pred in zip(truths, preds)])


def exact(truths, preds):
    return mean([compute_exact(truth, pred) for truth, pred in zip(truths, preds)])


def pearson_corr(preds, labels):
    return pearsonr(preds, labels)[0]


def spearman_corr(preds, labels):
    return spearmanr(preds, labels)[0]


def convert_BIO_labels(filename):
    result_labels = []
    with open(filename, "r", encoding="utf-8") as file:
        cnt = 0
        for line in file:
            line = re.sub(r"\*(\w+)", r"\1*", line)
            tokens = re.sub(
                r'[!"#$%&\'()+,-.:;<=>?@[\\\]^_`{\|}~⁇]', " ", line.strip()
            ).split()
            seq_label = []
            start_entity = 0
            entity_type = "O"
            for idx, token in enumerate(tokens):
                if token.endswith("*"):
                    start_entity += (
                        1 if (start_entity == 0 or token[:-1] != entity_type) else -1
                    )
                    entity_type = token[:-1]
                else:
                    if start_entity == 0:
                        seq_label.append("O")
                        entity_type = "O"
                    elif start_entity < 0:
                        raise "Something errors"
                    else:
                        if tokens[idx - 1].endswith("*"):
                            seq_label.append("B-" + entity_type.upper())
                        else:
                            seq_label.append("I-" + entity_type.upper())

            result_labels.append(seq_label)
            cnt += 1
    #             if cnt % 100 == 0:
    #                 print('Processed %d sentences' % cnt)
    return result_labels



import logging
import os
import re
import glob
import requests
import subprocess
import pandas as pd
from bs4 import BeautifulSoup
from datetime import datetime, timedelta
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class BOKMniutes:
    def __init__(self, **args):
        self.args = eKonf.to_config(args)
        # print(self.args)
        self.autoload = self.args.get('autoload', True)
        self.from_date = str(self.args.scrap.from_date)
        self.base_url = self.args.scrap.base_url
        self.url = self.args.scrap.url
        self.raw_hwp_dir = self.args.scrap.raw_hwp_dir
        self.raw_txt_dir = self.args.scrap.raw_txt_dir
        os.makedirs(self.raw_hwp_dir, exist_ok=True)
        os.makedirs(self.raw_txt_dir, exist_ok=True)
        self.file_prefix = self.args.scrap.file_prefix

        self.output_dir = self.args.output_dir
        self.output_file = os.path.join(self.output_dir, self.args.output_file)
        self.force_download = self.args.force_download

        user_agent = "Mozilla/5.0"
        self.headers = {"User-Agent": user_agent}
        self.minutes_urls = []
        self.minutes_hwp_files = []
        self.minutes_txt_files = []

        if self.autoload:
            self.build()

    def build(self):
        if not os.path.exists(self.output_file) or self.force_download:
            self.get_minutes_list()
            self.get_minutes_files()
            self.convert_hwp_to_txt()
            self.build_minutes()
        else:
            log.info(f"{self.output_file} already exists. skipping..")

    def build_minutes(self):

        docs = []
        for filepath in glob.glob(os.path.join(self.raw_txt_dir, "*.txt")):
            doc = preprocess_minutes(filepath, self.file_prefix)
            docs += doc

        df = pd.DataFrame(
            docs, columns=["id", "filename", "mdate", "rdate", "section", "text"]
        )
        df = df.dropna()

        if not self.force_download:
            if os.path.isfile(self.output_file):
                log.info("minutes file already exists. combining with the existing file..")
                minutes_df = pd.read_csv(self.output_file, index_col=None)
                df = minutes_df.combine_first(df)
                df = df.drop_duplicates(subset=["id"])

        df.to_csv(self.output_file, encoding="utf-8", index=False)
        print(df.tail())
        log.info(f"Saved {len(docs)} documents to {self.output_file}")

    def convert_hwp_to_txt(self):
        for filepath in glob.glob(os.path.join(self.raw_hwp_dir, "*.hwp")):
            filename = os.path.basename(filepath)
            hwp_file = f"{self.raw_hwp_dir}/{filename}"
            txt_file = f"{self.raw_txt_dir}/{filename[:-4]}.txt"
            self.minutes_txt_files.append(txt_file)

            if os.path.exists(txt_file):
                log.info(f"{txt_file} already exists")
                continue

            command = f"hwp5txt --output={txt_file} {hwp_file}"
            subprocess.Popen(command, shell=True)
            log.info(f"converted {hwp_file} to {txt_file}")

    def get_minutes_files(self):
        for info in self.minutes_urls:
            self.get_minutes_file(
                info["url"], info["meeting_date"], info["release_date"]
            )

    def get_minutes_file(self, page_addr, mdate, rdate):
        filename = (
            self.file_prefix
            + "_"
            + mdate.strftime("%Y%m%d")
            + "_"
            + rdate.strftime("%Y%m%d")
            + ".hwp"
        )
        filepath = f"{self.raw_hwp_dir}/{filename}"

        if os.path.exists(filepath):
            self.minutes_hwp_files.append(filepath)
            print(f"{filepath} already exists")
            return

        page = requests.get(page_addr)
        soup = BeautifulSoup(page.content, "html.parser")
        links = soup.find("div", class_="addfile").find_all("a")
        for link in links:
            link_filename = link.get_text()
            link_filename = (
                link_filename.replace("\r", "").replace("\t", "").replace("\n", "")
            )
            if link_filename[-3:] == "hwp":
                file_addr = self.base_url + link["href"]
                file_res = requests.get(file_addr)

                with open(filepath, "wb") as f:
                    f.write(file_res.content)
                self.minutes_hwp_files.append(filepath)
                print(f"saved {link_filename} as {filename} from {file_addr}")

    # Because this is for the recent minutes, changed the address of the list to that of the rss feed of the BOK minutes.
    def get_minutes_list(self):
        from_date = datetime.strptime(self.from_date, "%Y%m%d")
        url = self.url
        page = requests.get(url, headers=self.headers)

        soup = BeautifulSoup(page.content, "html.parser")
        brdList = soup.find_all("item")

        for post in brdList:
            pubdate = post.find("pubdate").get_text().strip()
            guid = post.find("guid").get_text().strip()
            title = post.find("title").get_text().strip()
            # description = post.find("description").get_text().strip()
            # if description.replace(' ','').find('통화정책방향') >= 0:
            mdate = title[title.find(")(") + 2 : -1]
            if mdate[-1] == ".":
                mdate = mdate[:-1]
            mdate = datetime.strptime(mdate, "%Y.%m.%d")
            if mdate < from_date:
                break

            rdate = pubdate[5:16]
            rdate = datetime.strptime(rdate, "%d %b %Y")
            info = {
                "meeting_date": mdate,
                "release_date": rdate,
                "url": guid,
                "title": title,
            }
            self.minutes_urls.append(info)


def tidy_sentences(section):
    sentence_enders = re.compile(r"((?<=[함음됨임봄짐움])(\s*\n|\.|;)|(?<=다)\.)\s*")
    splits = list((m.start(), m.end()) for m in re.finditer(sentence_enders, section))
    starts = [0] + [i[1] for i in splits]
    ends = [i[0] for i in splits]
    sentences = [section[start:end] for start, end in zip(starts[:-1], ends)]
    for i, s in enumerate(sentences):
        sentences[i] = (s.replace("\n", " ").replace(" ", " ")) + "."

    text = "\n".join(sentences) if len(sentences) > 0 else ""
    return sentences, text


def preprocess_minutes(filepath, file_prefix):
    filename = os.path.basename(filepath).replace(".txt", "")
    fileinfo = filename.split("_")
    mdate = datetime.strptime(fileinfo[1], "%Y%m%d") + timedelta(hours=10)
    rdate = datetime.strptime(fileinfo[2], "%Y%m%d") + timedelta(hours=16)
    fileinfo[0] = file_prefix
    filename = "_".join(fileinfo)
    doc = []

    print("open file: {}".format(filepath))
    minutes = open(filepath, encoding="utf-8").read()

    pos = re.search(
        "(.?국내외\s?경제\s?동향.?과 관련하여,?|\(가\).+경제전망.*|\(가\) 국내외 경제동향 및 평가)\n?\s*일부 위원은",
        minutes,
        re.MULTILINE,
    )
    s1 = pos.start() if pos else -1
    # pos = re.search('(.?외환.?국제금융\s?동향.?과 관련하여,?|\(나\) 외환.국제금융\s?(및 금융시장)?\s?동향)\n?\s*일부 위원은', minutes, re.MULTILINE)
    # pos = re.search('(.?외환.?국제금융\s?동향.?과 관련하여.*|\(나\) 외환.국제금융\s?(및 금융시장)?\s?동향)\n?\s*일부 위원은', minutes, re.MULTILINE)
    pos = re.search(
        "(.?외환.?국제금융\s?동향.?과 관련하여.*|\(나\) 외환.국제금융\s?(및 금융시장)?\s?동향)\n?\s*(일부 위원은|대부분의 위원들은)",
        minutes,
        re.MULTILINE,
    )
    s2 = pos.start() if pos else -1
    pos = re.search(
        "(.?금융시장\s?동향.?과 관련하여,?|\(다\) 금융시장\s?동향)\n?\s*일부 위원은", minutes, re.MULTILINE
    )
    s3 = pos.start() if pos else -1
    # pos = re.search('((\((다|라)\) )?.?통화정책방향.?에 관한 토론,?|이상과 같은 의견교환을 바탕으로.*통화정책방향.*에 관해 다음과 같은 토론이 있었음.*)\n?', minutes, re.MULTILINE)
    # pos = re.search('((\((다|라)\) )?.?통화정책.?방향.?에 관한 토론,?|이상과 같은 의견교환을 바탕으로.*통화정책방향.*에.*토론.*)\n?', minutes, re.MULTILINE)
    # pos = re.search('((\((다|라)\) )?.?통화정책\s?방향.?에 관한 토론,?|이상과 같은 의견교환을 바탕으로.*통화정책\s?방향.*에.*토론.*)\n?', minutes, re.MULTILINE)
    pos = re.search(
        "((\((다|라)\) )?.?통화정책\s?방향.?에 관한 토론,?|이상과 같은 의견\s?교환을 바탕으로.*통화정책\s?방향.*에.*토론.*)\n?",
        minutes,
        re.MULTILINE,
    )
    s4 = pos.start() if pos else -1
    pos = re.search("(\(4\) 정부측 열석자 발언.*)\n?", minutes, re.MULTILINE)
    s5 = pos.start() if pos else -1
    pos = re.search(
        "(\(.*\) 한국은행 기준금리 결정에 관한 위원별 의견\s?개진|이상과 같은 토론에 이어 .* 관한 위원별 의견개진이 있었음.*)\n?",
        minutes,
        re.MULTILINE,
    )
    s6 = pos.start() if pos else -1
    # pos = re.search('(\(\s?.*\s?\) ()(심의결과|토의결론))\n?', minutes, re.MULTILINE)
    # s7 = pos.start() if pos else -1
    positer = re.finditer("(\(\s?.*\s?\) ()(심의결과|토의결론))\n?", minutes, re.MULTILINE)
    s7 = [pos.start() for pos in positer if pos.start() > s6]
    s7 = s7[0] if s7 else -1

    # 국내외 경제동향
    bos = s1
    eos = s2
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("(일부|대부분의) 위원들?은", section, re.MULTILINE)
    bos = pos.start() if pos else -1
    section = section[bos:] if bos >= 0 else section
    section1, section1_txt = tidy_sentences(section)

    # 외환․국제금융 동향
    bos = s2
    eos = s3 if s3 >= 0 else s4
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("(일부|대부분의) 위원들?은", section, re.MULTILINE)
    bos = pos.start() if pos else -1
    section = section[bos:] if bos >= 0 else section
    section2, section2_txt = tidy_sentences(section)

    # 금융시장 동향
    bos = s3
    eos = s4
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("(일부|대부분의) 위원들?은", section, re.MULTILINE)
    bos = pos.start() if pos else -1
    section = section[bos:] if bos >= 0 else section
    section3, section3_txt = tidy_sentences(section)

    # 통화정책방향
    bos = s4
    eos = s5 if s5 >= 0 else s6 if s6 >= 0 else s7
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("(일부|대부분의) 위원들?은", section, re.MULTILINE)
    bos = pos.start() if pos else -1
    section = section[bos:] if bos >= 0 else section
    section4, section4_txt = tidy_sentences(section)

    # 위원별 의견 개진
    bos = s6
    eos = s7
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("(일부|대부분의) 위원들?은", section, re.MULTILINE)
    bos = pos.start() if pos else -1
    section = section[bos:] if bos >= 0 else section
    section5, section5_txt = tidy_sentences(section)

    # 정부측 열석자 발언
    bos = s5
    eos = s6
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("정부측 열석자 발언", section, re.MULTILINE)
    bos = pos.end() + 1 if pos else -1
    section = section[bos:] if bos >= 0 else section
    section6, section6_txt = tidy_sentences(section)

    sections = [
        "Economic Situation",
        "Foreign Currency",
        "Financial Markets",
        "Monetary Policy",
        "Participants’ Views",
        "Government’s View",
    ]
    # section_texts = (section1, section2, section3, section4, section5, section6)
    section_texts = (
        section1_txt,
        section2_txt,
        section3_txt,
        section4_txt,
        section5_txt,
        section6_txt,
    )

    print(" ==> text processing completed: {}".format(filename))

    if any(section_texts):
        for s, (section, text) in enumerate(zip(sections, section_texts)):
            # for p, text in enumerate(sentences):
            id = f"{filename}_S{s+1}"
            # id = f'{filename}_S{s+1}_P{p+1}'
            # row = (id, filename, mdate, rdate, section, p, text)
            row = (id, filename, mdate, rdate, section, text)
            doc.append(row)
    else:
        print("Empty!")

    return doc



import numpy as np
import pandas as pd
from ekorpkit import eKonf
from ekorpkit.io.file import load_dataframe


class Lexicon:
    def __init__(self, **args) -> None:
        args = eKonf.to_dict(args)
        self.args = args
        self.name = args["name"]
        self.fullname = args["fullname"]
        self.source = args.get("source", None)
        self.lang = args.get("lang", None)
        self._lexicon_ngram_delim = args.get("ngram_delim")
        self._lexicon_ngram_max_n = args.get("ngram_max_n")
        self._lexicon_features = args["lexicon_features"]
        self._lexicon_path = args["lexicon_path"]
        self._word_key = args.get("word_key", "word")
        self._arr_word_key = "_" + self._word_key
        self._ngram_len_key = "_ngram_len"
        self._token_count_key = "count"
        self._ignore_pos = self.args.get("ignore_pos", False)
        self._postag_length = self.args.get("postag_length")
        if not self._postag_length:
            self._postag_length = 1
        self._postag_delim = self.args.get("postag_delim")
        if not self._postag_delim:
            self._postag_delim = "/"
        self._lexicon = None
        self._analyze_args = args.get("analyze", {})
        self._features = self._analyze_args.get("features", None)
        self._ngram_distiance_tolerance = self._analyze_args.get(
            "ngram_distiance_tolerance", 0
        )
        self._ngram_delim = self._analyze_args.get("ngram_delim")

        self._load_lexicon()

    def _load_lexicon(self, lexicon_path=None):
        lexicon_path = lexicon_path or self._lexicon_path
        df = load_dataframe(lexicon_path)
        df = df.dropna(subset=[self._word_key])
        df[self._word_key] = df[self._word_key].apply(self._prepare_token)
        if self._lexicon_ngram_delim:
            df[self._arr_word_key] = df[self._word_key].str.split(
                self._lexicon_ngram_delim
            )
            df[self._ngram_len_key] = df[self._arr_word_key].apply(len)
            df.sort_values(by=self._ngram_len_key, ascending=False, inplace=True)
            self._lexicon = df[
                [self._word_key, self._arr_word_key, self._ngram_len_key]
                + self._lexicon_features
            ]
        else:
            self._lexicon = df[[self._word_key] + self._lexicon_features]

    def _prepare_token(self, token, ngram_delim=None, strip_pos=None, as_list=False):
        ngram_delim = ngram_delim or self._ngram_delim
        strip_pos = strip_pos or self._ignore_pos
        token = token.lower()

        def _get_word(token):
            token_pos = token.split(self._postag_delim)
            if strip_pos:
                return token_pos[0]
            return (
                token
                if len(token_pos) == 1
                else token_pos[0]
                + self._postag_delim
                + token_pos[1][: self._postag_length]
            )

        if ngram_delim:
            _token = [_get_word(subtoken) for subtoken in token.split(ngram_delim)]
            return _token if as_list else ngram_delim.join(_token)
        else:
            return _get_word(token)

    def _prepare_tokens(
        self, tokens, ngram_delim=None, strip_pos=None, ngram_distiance_tolerance=None
    ):
        ngram_delim = ngram_delim or self._ngram_delim
        strip_pos = strip_pos or self._ignore_pos
        ngram_distiance_tolerance = (
            ngram_distiance_tolerance or self._ngram_distiance_tolerance
        )

        token_dict = {}
        ngrams_dict = {}
        for token in tokens:
            token = self._prepare_token(token, ngram_delim, strip_pos, False)
            if token in token_dict:
                token_dict[token][self._token_count_key] += 1
            else:
                token_dict[token] = {self._token_count_key: 1}
            if ngram_distiance_tolerance > 0 and ngram_delim:
                ngram = self._prepare_token(token, ngram_delim, strip_pos, True)
                if len(ngram) > 1:
                    ngrams_dict[token] = ngram
        return token_dict, ngrams_dict

    @property
    def n_features(self):
        return len(self._lexicon_features)

    def __len__(self):
        return len(self._lexicon)

    def __getitem__(self, key):
        return self._lexicon.loc[key]

    def __contains__(self, key):
        return key in self._lexicon

    def __iter__(self):
        return iter(self._lexicon)

    def __str__(self):
        classname = self.__class__.__name__
        s = f"{classname}\n----------\n"
        s += f"name: {self.name}\n"
        s += f"fullname: {self.fullname}\n"
        s += f"source: {self.source}\n"
        s += f"lang: {self.lang}\n"
        s += f"lexicon_features: {self._lexicon_features}\n"
        s += f"lexicon_path: {self._lexicon_path}\n"
        s += f"word_key: {self._word_key}\n"
        s += f"ignore_pos: {self._ignore_pos}\n"
        return s

    @staticmethod
    def check_subset_tokens(word, ngrams, ngram_distance_tolerance=0):
        for token, ngram in ngrams.items():
            if (
                set(word).issubset(set(ngram))
                and len(ngram) >= len(word)
                and len(ngram) - len(word) <= ngram_distance_tolerance
            ):
                return token
        return np.nan

    def analyze(
        self,
        tokens: list,
        features=None,
        ngram_distiance_tolerance=None,
        ngram_delim=None,
        **kwargs,
    ):
        features = features or self._features or self._lexicon_features
        ignore_pos = self._ignore_pos
        ngram_distiance_tolerance = (
            ngram_distiance_tolerance or self._ngram_distiance_tolerance
        )
        ngram_delim = ngram_delim or self._ngram_delim

        token_dict, ngrams_dict = self._prepare_tokens(
            tokens, ngram_delim, ignore_pos, ngram_distiance_tolerance
        )
        df = self._lexicon
        if len(ngrams_dict) > 0:
            df_token = df[df[self._ngram_len_key] == 1]
            df_ngram = df[df[self._ngram_len_key] > 1]
        else:
            df_token = df
            df_ngram = None
        token_features = df_token[df_token[self._word_key].isin(token_dict.keys())]
        if df_ngram is not None:
            check = df_ngram[self._arr_word_key].apply(
                lambda x: self.check_subset_tokens(
                    x, ngrams_dict, ngram_distiance_tolerance
                )
            )
            ngram_features = df_ngram.loc[list(check.dropna().index), :]
            if not ngram_features.empty:
                ngram_features[self._word_key] = check
                token_features = pd.concat(
                    [token_features, ngram_features], axis=0
                ).drop_duplicates(subset=self._word_key, keep="first")

        feature_dict = token_features.set_index(self._word_key)[features].to_dict(
            "index"
        )

        return_dict = {}
        for token, features in feature_dict.items():
            if not isinstance(token, str):
                continue
            features.update(token_dict[token])
            return_dict[token] = features
        return return_dict



import logging
import os
import pandas as pd
from glob import glob
from pathlib import Path
from ekorpkit.utils.func import elapsed_timer
from ekorpkit import eKonf

log = logging.getLogger(__name__)


def get_filepaths(
    filename_patterns, base_dir=None, recursive=True, verbose=True, **kwargs
):
    if isinstance(filename_patterns, str):
        filename_patterns = [filename_patterns]
    filepaths = []
    for file in filename_patterns:
        file = os.path.join(base_dir, file) if base_dir else file
        if os.path.exists(file):
            if Path(file).is_file():
                filepaths.append(file)
        else:
            filepaths += glob(file, recursive=recursive)
    filepaths = [fp for fp in filepaths if Path(fp).is_file()]
    if verbose:
        log.info(f"Processing [{len(filepaths)}] files from [{filename_patterns}]")

    return filepaths


def get_files_from_archive(archive_path, filetype=None):
    import tarfile
    from zipfile import ZipFile

    if ".tar.gz" in archive_path:
        log.info(f"::Extracting files from {archive_path} with tar.gz")
        archive_handle = tarfile.open(archive_path, "r:gz")
        files = [
            (file, file.name) for file in archive_handle.getmembers() if file.isfile()
        ]
        open_func = archive_handle.extractfile
    elif ".tar.bz2" in archive_path:
        log.info(f"::Extracting files from {archive_path} with tar.bz2")
        archive_handle = tarfile.open(archive_path, "r:bz2")
        files = [
            (file, file.name) for file in archive_handle.getmembers() if file.isfile()
        ]
        open_func = archive_handle.extractfile
    elif ".zip" in archive_path:
        log.info(f"::Extracting files from {archive_path} with zip")
        archive_handle = ZipFile(archive_path)
        files = [
            (file, file.encode("cp437").decode("euc-kr"))
            for file in archive_handle.namelist()
        ]
        open_func = archive_handle.open
    else:
        # print(f'::{archive_path} is not archive, use generic method')
        files = [(archive_path, os.path.basename(archive_path))]
        archive_handle = None
        open_func = None
    if filetype:
        files = [file for file in files if filetype in file[1]]

    return files, archive_handle, open_func


def save_dataframe(
    df,
    filepath=None,
    filetype=None,
    verbose=False,
    index=False,
    columns=None,
    name=None,
    _name_=None,
    output_dir=None,
    output_file=None,
    **kwargs,
):
    if df is None:
        log.warning("Dataframe is None")
        return df
    if isinstance(columns, list):
        df = df[columns]
    if verbose > 1:
        print(df.tail())

    if filepath:
        output_dir = os.path.dirname(filepath)
        output_file = os.path.basename(filepath)
    if output_file:
        fileinfo = os.path.splitext(output_file)
        filename = fileinfo[0]
        filetype = (
            fileinfo[1]
            if len(fileinfo) > 1
            else ("parquet" if not filetype else filetype)
        )
    else:
        filename = f"{name}"
        if not filetype:
            filetype = "parquet"
    filetype = "." + filetype.replace(".", "")
    if _name_ is not None:
        if _name_.endswith(filetype):
            filename = f"{filename}-{_name_}"
        else:
            filename = f"{filename}-{_name_}{filetype}"
    else:
        filename = f"{filename}{filetype}"
    filepath = os.path.join(output_dir, filename)
    os.makedirs(os.path.abspath(output_dir), exist_ok=True)

    log.info(f"Saving dataframe as {filepath}")
    with elapsed_timer(format_time=True) as elapsed:
        if "csv" in filetype or "tsv" in filetype:
            df.to_csv(filepath, index=index)
        elif "parquet" in filetype:
            df.to_parquet(filepath, compression="gzip", engine="pyarrow")
        else:
            raise ValueError("filetype must be .csv or .parquet")
        if verbose:
            log.info(" >> elapsed time to save data: {}".format(elapsed()))
    if verbose:
        log.info(f" >> saved dataframe to {filepath}")


def concat_dataframes(
    data,
    add_key_as_name=False,
    name_column=eKonf.Keys.NAME,
    concat={},
    columns=None,
    verbose=False,
    **kwargs,
):
    if isinstance(data, dict):
        log.info(f"Concatenating {len(data)} dataframes")
        dfs = []
        for df_name in data:
            df_each = data[df_name]
            if isinstance(columns, list):
                columns = [c for c in columns if c in df_each.columns]
                df_each = df_each[columns]
            if add_key_as_name:
                df_each[name_column] = df_name
            dfs.append(df_each)
        data = pd.concat(dfs, **concat)
        return data
    else:
        if verbose:
            log.info("Returning the original dataframe")
        return data


def load_dataframe(
    filepath=None,
    filetype=None,
    verbose=False,
    index_col=None,
    columns=None,
    name=None,
    data_dir=None,
    data_file=None,
    **kwargs,
):
    if filepath:
        data_dir = os.path.dirname(filepath)
        data_file = os.path.basename(filepath)
    if data_file:
        fileinfo = os.path.splitext(data_file)
        filename = fileinfo[0]
        filetype = (
            fileinfo[1]
            if len(fileinfo) > 1
            else ("parquet" if not filetype else filetype)
        )
    else:
        filename = f"{name}"
        if not filetype:
            filetype = "parquet"
    filetype = "." + filetype.replace(".", "")
    filename = f"{filename}{filetype}"
    filepath = os.path.join(data_dir, filename)

    if not os.path.exists(filepath):
        log.warning(f"File {filepath} does not exist")
        return None
    log.info("Loading data from {}".format(filepath))
    with elapsed_timer(format_time=True) as elapsed:
        if "csv" in filetype:
            df = pd.read_csv(filepath, index_col=index_col, **kwargs)
        elif "tsv" in filetype:
            df = pd.read_csv(filepath, index_col=index_col, sep="\t", **kwargs)
        elif "parquet" in filetype:
            df = pd.read_parquet(filepath, engine="pyarrow")
        else:
            raise ValueError("filetype must be .csv or .parquet")
        if isinstance(columns, list):
            columns = [c for c in columns if c in df.columns]
            df = df[columns]
        if verbose:
            log.info(" >> elapsed time to load data: {}".format(elapsed()))
    return df



import logging
import pandas as pd
from .base import BaseSentimentAnalyser

log = logging.getLogger(__name__)


class HIV4SA(BaseSentimentAnalyser):
    """
    A class for sentiment analysis using the HIV4 lexicon.
    """

    def __init__(self, preprocessor=None, lexicon=None, **kwargs):
        super().__init__(preprocessor=preprocessor, lexicon=lexicon, **kwargs)

    def _get_score(self, tokens, lexicon_features, feature="polarity"):
        """Get score for features.

        :returns: int
        """
        lxfeat_names = self._features.get(feature).get("lexicon_features")
        lxfeat = pd.DataFrame.from_dict(lexicon_features, orient="index")
        score = {}
        if feature == "polarity":
            lxfeat["pos"] = lxfeat.apply(
                lambda x: 1 * x["count"] if x["Positiv"] else 0, axis=1
            )
            lxfeat["neg"] = lxfeat.apply(
                lambda x: 1 * x["count"] if x["Negativ"] else 0, axis=1
            )
            lxfeat_agg = lxfeat.agg({"pos": "sum", "neg": "sum"})
            polarity = (lxfeat_agg["pos"] - lxfeat_agg["neg"]) / (
                lxfeat_agg["pos"] + lxfeat_agg["neg"] + self.EPSILON
            )
            subjectivity = (lxfeat_agg["pos"] + lxfeat_agg["neg"]) / (
                len(tokens) + self.EPSILON
            )
            score["polarity"] = polarity
            score["subjectivity"] = subjectivity
        elif isinstance(lxfeat_names, str):
            lxfeat[feature] = lxfeat.apply(
                lambda x: 1 * x["count"] if x[lxfeat_names] else 0, axis=1
            )
            lxfeat_agg = lxfeat.agg({feature: "sum"})
            feat_score = lxfeat_agg[feature] / (len(tokens) + self.EPSILON)
            score[feature] = feat_score

        return score

    def _assign_class(self, score, feature="polarity"):
        """Assign class to a score.

        :returns: str
        """
        labels = self._features.get(feature).get("labels")
        if labels:
            score["label"] = ""
            for label, thresh in labels.items():
                if isinstance(thresh, str):
                    thresh = eval(thresh)
                if score[feature] >= thresh[0] and score[feature] <= thresh[1]:
                    score["label"] = label
        return score


class LMSA(BaseSentimentAnalyser):
    """
    A class for sentiment analysis using the LM lexicon.
    """

    def __init__(self, preprocessor=None, lexicon=None, **kwargs):
        super().__init__(preprocessor=preprocessor, lexicon=lexicon, **kwargs)

    def _get_score(self, tokens, lexicon_features, feature="polarity"):
        """Get score for features.

        :returns: int
        """
        lxfeat_names = self._features.get(feature).get("lexicon_features")
        lxfeat = pd.DataFrame.from_dict(lexicon_features, orient="index")
        score = {}
        if feature == "polarity":
            lxfeat["pos"] = lxfeat.apply(
                lambda x: 1 * x["count"] if x["Positive"] > 0 else 0, axis=1
            )
            lxfeat["neg"] = lxfeat.apply(
                lambda x: 1 * x["count"] if x["Negative"] > 0 else 0, axis=1
            )
            lxfeat_agg = lxfeat.agg({"pos": "sum", "neg": "sum"})
            polarity = (lxfeat_agg["pos"] - lxfeat_agg["neg"]) / (
                lxfeat_agg["pos"] + lxfeat_agg["neg"] + self.EPSILON
            )
            subjectivity = (lxfeat_agg["pos"] + lxfeat_agg["neg"]) / (
                len(tokens) + self.EPSILON
            )
            score["positive"] = lxfeat_agg["pos"] / (len(tokens) + self.EPSILON)
            score["negative"] = lxfeat_agg["neg"] / (len(tokens) + self.EPSILON)
            score["num_tokens"] = len(tokens)
            score[feature] = polarity
            score["subjectivity"] = subjectivity
        elif isinstance(lxfeat_names, str):
            lxfeat[feature] = lxfeat.apply(
                lambda x: 1 * x["count"] if x[lxfeat_names] > 0 else 0, axis=1
            )
            lxfeat_agg = lxfeat.agg({feature: "sum"})
            feat_score = lxfeat_agg[feature] / (len(tokens) + self.EPSILON)
            score[feature] = feat_score

        return score

    def _assign_class(self, score, feature="polarity"):
        """Assign class to a score.

        :returns: str
        """
        label_key = feature + "_label"
        labels = self._features.get(feature).get("labels")
        if labels:
            score[label_key] = ""
            for label, thresh in labels.items():
                if isinstance(thresh, str):
                    thresh = eval(thresh)
                if score[feature] >= thresh[0] and score[feature] <= thresh[1]:
                    score[label_key] = label
        return score



import logging
from abc import ABCMeta, abstractmethod
from ekorpkit import eKonf

log = logging.getLogger(__name__)


class BaseSentimentAnalyser:
    __metaclass__ = ABCMeta
    """
    A base class for sentiment analysis.
    """

    EPSILON = 1e-6

    def __init__(self, preprocessor=None, lexicon=None, **kwargs):
        self._predict = kwargs.get("predict") or {}
        self._predict_features = self._predict.get("features")
        self._predict_features = self._predict_features or "polarity"
        self._eval = kwargs.get("eval") or {}
        self._features = kwargs.get("features") or {}
        self.verbose = kwargs.get("verbose", False)

        self._tokenizer = preprocessor["tokenizer"]
        if eKonf.is_instantiatable(self._tokenizer):
            if self.verbose:
                log.info(f"instantiating {self._tokenizer['_target_']}...")
            self._tokenizer = eKonf.instantiate(self._tokenizer)

        self._lexicon = lexicon
        if eKonf.is_instantiatable(self._lexicon):
            if self.verbose:
                log.info(f"instantiating {self._lexicon['_target_']}...")
            self._lexicon = eKonf.instantiate(self._lexicon)

    def tokenize(self, text):
        """
        :type text: str
        :returns: list
        """
        return self._tokenizer.tokenize(text)

    def analyze(self, text, **kwargs):
        """
        :type text: str
        :returns: list
        """
        return self._lexicon.analyze(self.tokenize(text), **kwargs)

    @abstractmethod
    def _get_score(self, tokens, lexicon_features, feature="polarity"):
        """Get score for features.

        :returns: int
        """
        raise NotImplementedError("Must override segment")

    @abstractmethod
    def _assign_class(self, score, feature="polarity"):
        """Assign class to a score.

        :returns: str
        """
        raise NotImplementedError("Must override segment")

    def predict(self, text, features=None):
        """Get score for a list of terms.

        :returns: dict
        """
        features = features or self._predict_features
        if isinstance(features, str):
            features = [features]
        tokens = self.tokenize(text)
        lex_feats = self._predict.get("lexicon_features")
        lexicon_features = self._lexicon.analyze(tokens, features=lex_feats)
        scores = {}
        for feather in features:
            score = self._get_score(tokens, lexicon_features, feature=feather)
            score = self._assign_class(score, feature=feather)
            scores.update(score)

        return scores



"""Guesslang machine learning model"""

import json
import logging
from pathlib import Path
from statistics import mean, stdev
from tempfile import TemporaryDirectory
from typing import List, Tuple, Optional

from . import model


log = logging.getLogger(__name__)

DATA_DIR = Path(__file__).absolute().parent.joinpath("data")
DEFAULT_MODEL_DIR = DATA_DIR.joinpath("model")
LANGUAGES_FILE = DATA_DIR.joinpath("languages.json")
TEST_REPORT_FILE = "test-report.json"


class Guess:
    """Guess the programming language of a source code.

    :param model_dir: Guesslang machine learning model directory.
    """

    def __init__(
        self, model_dir: Optional[str] = None, verbose: Optional[int] = 0
    ) -> None:
        if model_dir:
            self._saved_model_dir = model_dir
        else:
            self._saved_model_dir = str(DEFAULT_MODEL_DIR)

        try:
            self._model = model.load(self._saved_model_dir)
        except OSError:
            self._model = None

        language_json = LANGUAGES_FILE.read_text()
        self._language_map = json.loads(language_json)
        self._extension_map = {ext: name for name, ext in self._language_map.items()}
        self.verbose = verbose

    @property
    def is_trained(self) -> bool:
        """Check if the current machine learning model is trained.
        Only trained models can be used for prediction.

        :return: the model training status.
        """
        return self._model is not None

    @property
    def supported_languages(self) -> List[str]:
        """List supported programming languages

        :return: language name list.
        """
        return list(self._language_map)

    def language_name(self, source_code: str) -> Optional[str]:
        """Predict the programming language name of the given source code.

        :param source_code: source code.
        :return: the language name
            or ``None`` if no programming language is detected.
        """
        if not source_code.strip():
            log.warning("Empty source code provided")
            return None

        language_probabilities = self.probabilities(source_code)
        probabilities = [value for _, value in language_probabilities]
        if not self._is_reliable(probabilities):
            if self.verbose > 5:
                log.warning("No programming language detected")
            return None

        language_name, _ = language_probabilities[0]
        return language_name

    def probabilities(self, source_code: str) -> List[Tuple[str, float]]:
        """Gives the probability that the source code is written
        in each of the supported languages.

        The probabilities are sorted from the most to the least probable
        programming language.

        :param source_code: source code.
        :return: list of language names associated with their probability.
        """
        if not self.is_trained:
            log.error("Cannot predict using an untrained model")
            raise GuesslangError(
                f"Cannot predict using the untrained model located at "
                f"{self._saved_model_dir}. "
                f"Train your model with `guess.train(source_files_dir)`"
            )

        return model.predict(self._model, self._extension_map, source_code)

    def train(self, source_files_dir: str, max_steps: int) -> float:
        """Train guesslang to recognize programming languages.

        The machine learning model is trained from source code files.
        The files should be split in three subdirectories named
        "train", "valid" and "test".

        :raise GuesslangError: when the training cannot be run.
        :param source_files_dir: directory that contains
            the "train", "valid" and "test" datasets.
        :return: training accuracy, a value between 0 and 1.
        """

        log.debug("Run safety checks before starting the training")

        if self.is_trained:
            log.error("Model already trained")
            raise GuesslangError(
                f"The current model located at {self._saved_model_dir} "
                f"is already trained"
            )

        input_path = Path(source_files_dir)
        for dirname in model.DATASET.values():
            dataset_path = input_path.joinpath(dirname)
            if not dataset_path.is_dir():
                log.error(f"Dataset directory missing {dataset_path}")
                raise GuesslangError(f"No dataset directory: {dataset_path}")

        log.debug("Run the training")
        extensions = list(self._extension_map)
        with TemporaryDirectory() as model_logs_dir:
            estimator = model.build(model_logs_dir, extensions)
            metrics = model.train(estimator, source_files_dir, max_steps)
            log.info(f"Training metrics: {metrics}")
            model.save(estimator, self._saved_model_dir)

        log.debug(f"Test newly trained model {self._saved_model_dir}")
        self._model = model.load(self._saved_model_dir)
        matches = model.test(self._model, source_files_dir, self._extension_map)

        report_file = Path(self._saved_model_dir).joinpath(TEST_REPORT_FILE)
        json_data = json.dumps(matches, indent=2, sort_keys=True)
        report_file.write_text(json_data)
        log.debug(f"Test report stored in {report_file}")

        languages = self._language_map.keys()
        total = sum(sum(values.values()) for values in matches.values())
        success = sum(matches[language][language] for language in languages)
        accuracy = success / total
        log.debug(f"Accuracy = {success} / {total} = {accuracy:.2%}")
        return accuracy

    @staticmethod
    def _is_reliable(probabilities: List[float]) -> bool:
        """Arbitrary rule to determine if the prediction is reliable:

        The predicted language probability must be higher than
        2 standard deviations from the mean.
        """
        threshold = mean(probabilities) + 2 * stdev(probabilities)
        predicted_language_probability = max(probabilities)
        return predicted_language_probability > threshold


class GuesslangError(Exception):
    """Guesslang exception class"""



import logging
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from cmath import isinf
from pathlib import Path
from .base import set_style, set_figure
from .classification import confusion_matrix
from ekorpkit import eKonf


log = logging.getLogger(__name__)


def plot(data, verbose=False, **kwargs):
    kwargs = eKonf.to_dict(kwargs)
    plots = kwargs.get("plots")
    figures = kwargs.get("figures")
    series = {} or kwargs.get("series")
    plot = {} or kwargs.get("plot")
    subplots = {} or kwargs.get("subplots")
    # figure = {} or kwargs.get("figure")
    savefig = {} or kwargs.get("savefig")

    if data is None:
        log.warning("No data to plot")
        return

    set_style(**plot)
    figsize = plot.get("figsize", None)
    if figsize is not None and isinstance(figsize, str):
        figsize = eval(figsize)

    dataform = series.get("form", "individual")
    xcol = series["x"]
    if isinstance(xcol, list):
        xcol = xcol[0]
    if isinstance(data, pd.DataFrame):
        x = data[xcol] if xcol in data.columns else data.index
    else:
        x = xcol
    ycols = series["y"]
    if isinstance(ycols, str):
        ycols = [ycols]

    sencondary_axes = {}
    if subplots["nrows"] > 1 or subplots["ncols"] > 1:
        fig, axes = plt.subplots(**subplots, figsize=figsize)
        ax = None
        axes = axes.flatten()
    else:
        plt.figure(figsize=figsize, tight_layout=True)
        ax = plt.gca()
        axes = [ax]

    if dataform == "individual":
        if isinstance(plots, dict):
            plots = [plots]
        for _plot_cfg_ in plots:
            _func_ = eval(_plot_cfg_.get(eKonf.Keys.FUNC))
            _x = _plot_cfg_.pop("x") or x
            _y = _plot_cfg_.pop("y")
            secondary_y = _plot_cfg_.get("secondary_y", False)
            if secondary_y:
                secondary_to = _plot_cfg_.get("secondary_to", 0)
                log.info(f"Creating secondary axis to axis[{secondary_to}]")
                if secondary_to in sencondary_axes:
                    ax = sencondary_axes[secondary_to]
                else:
                    ax = axes[secondary_to].twinx()
                    sencondary_axes[secondary_to] = ax
            else:
                axno = _plot_cfg_.get("axno", 0)
                ax = axes[axno]
            _func_(ax, _x, _y, data, **_plot_cfg_)
    else:
        if isinstance(plots, list):
            _plot_cfg_ = plots[0]
        else:
            _plot_cfg_ = plots
        _func_ = eval(_plot_cfg_.get(eKonf.Keys.FUNC))
        if ycols and len(ycols) == 1:
            ycols = ycols[0]
        _x = _plot_cfg_.pop("x") or xcol
        _y = _plot_cfg_.pop("y") or ycols
        if _plot_cfg_.get("dataform"):
            dataform = _plot_cfg_.pop("dataform")
        if dataform == "wide":
            if isinstance(data, pd.DataFrame):
                if _x and _x in data.columns:
                    _data = data.set_index(_x)[_y]
                else:
                    _data = data[_y]
            else:
                _data = data
            _func_(ax, x=None, y=None, data=_data, **_plot_cfg_)
        else:
            _func_(ax, x=_x, y=_y, data=data, **_plot_cfg_)

    if figures is None:
        figures = []
    elif not isinstance(figures, (dict, list)):
        figures = []
    elif isinstance(figures, dict):
        figures = [figures]
    for _fig_cfg_ in figures:
        secondary_y = _fig_cfg_.get("secondary_y", False)
        if secondary_y:
            secondary_to = _fig_cfg_.get("secondary_to", 0)
            if secondary_to in sencondary_axes:
                ax = sencondary_axes[secondary_to]
            else:
                ax = None
        else:
            axno = _fig_cfg_.get("axno", 0)
            ax = axes[axno]
        add_decorations(ax, **_fig_cfg_)
        set_figure(ax, **_fig_cfg_)

    fname = savefig.get("fname", None)
    if fname:
        Path(fname).parent.mkdir(parents=True, exist_ok=True)
        plt.savefig(**savefig)
        log.info(f"Saved figure to {fname}")


def lineplot(ax=None, x=None, y=None, data=None, **kwargs):
    _parms_ = {} or kwargs.get(eKonf.Keys.PARMS)
    if ax is None:
        ax = plt.gca()
    sns.lineplot(x=x, y=y, data=data, ax=ax, **_parms_)


def stackplot(ax=None, x=None, y=None, data=None, **kwargs):
    _parms_ = {} or kwargs.get(eKonf.Keys.PARMS)
    if ax is None:
        ax = plt.gca()
    if x is None:
        x = data.index
    elif isinstance(x, str):
        x = data[x]
    plt.stackplot(x, data[y].T, **_parms_)


def scatter(ax=None, x=None, y=None, data=None, **kwargs):
    _parms_ = {} or kwargs.get(eKonf.Keys.PARMS)
    if ax is None:
        ax = plt.gca()
    sns.scatterplot(x=x, y=y, data=data, ax=ax, **_parms_)


def add_decorations(ax=None, **kwargs):
    if ax is None:
        ax = plt.gca()
    axvspans = [] or kwargs.get("axvspans")
    annotations = [] or kwargs.get("annotations")

    if isinstance(axvspans, dict):
        axvspans = [axvspans]
    if isinstance(annotations, dict):
        annotations = [annotations]

    for span in axvspans:
        if isinstance(span, dict):
            ax.axvspan(**span)
    for annot in annotations:
        if isinstance(annot, dict):
            x = annot.pop("x")
            y = annot.pop("y")
            if x and y:
                annot["xy"] = (x, y)
            xtext = annot.pop("xtext")
            ytext = annot.pop("ytext")
            if xtext and ytext:
                annot["xytext"] = (xtext, ytext)
            ax.annotate(**annot)


def treemap(
    df,
    fig_filepath,
    scale=1.5,
    columns=None,
    treemap=None,
    layout=None,
    update_layout=None,
    **kwargs,
):
    # textinfo = "label+value+percent parent+percent root"
    # import plotly.express as px
    import plotly.graph_objects as go

    labels = list(df[columns.label].to_list())
    values = list(df[columns.value].to_list())
    parents = list(df[columns.parent].to_list())

    layout = go.Layout(**layout)
    fig = go.Figure(
        go.Treemap(
            labels=labels,
            parents=parents,
            values=values,
            **treemap,
        ),
        layout=layout,
    )
    fig.update_layout(
        **update_layout,
    )

    fig.write_image(fig_filepath, scale=scale)


def barplot(df, columns=None, savefig={}, plot={}, figure={}, verbose=False, **kwargs):
    if df is None:
        if verbose:
            print("No data to plot")
        return
    set_style(**plot)
    figsize = plot.get("figsize", None)
    if figsize is not None and isinstance(figsize, str):
        figsize = eval(figsize)
    ycols = columns.yvalue
    xcol = columns.xvalue
    index = columns.get("index", None)
    if index:
        df.index = list(index)
        xcol = None
    data = df

    # plt.figure(figsize=figsize, tight_layout=True)
    stacked = plot.get("stacked", False)
    ax = data.plot(x=xcol, y=ycols, kind="bar", stacked=stacked, figsize=figsize)
    set_figure(ax, **figure)
    plt.tight_layout()

    fname = savefig.get("fname", None)
    if fname:
        Path(fname).parent.mkdir(parents=True, exist_ok=True)
        plt.savefig(**savefig)
        if verbose:
            print(f"Saved figure to {fname}")



import logging
from ekorpkit import eKonf
from ekorpkit.utils.func import elapsed_timer
from ekorpkit.pipelines.pipe import apply_pipeline


log = logging.getLogger(__name__)


def topic_tasks(**cfg):
    args = eKonf.to_config(cfg)
    corpus_cfg = args.corpus
    model_cfg = args.model.topic
    subtasks = args.task.topic
    _subtasks_ = args.task.topic._subtasks_

    with elapsed_timer(format_time=True) as elapsed:
        model = eKonf.instantiate(model_cfg)
        model.corpora = eKonf.instantiate(corpus_cfg)

        for subtask in _subtasks_:
            subtask_cfg = subtasks[subtask]
            if eKonf.Keys.TARGET in subtask_cfg:
                log.info(f"Instantiate {subtask} ...")
                eKonf.instantiate(subtask_cfg, model=model)
            elif eKonf.Keys.NAME in subtask_cfg:
                log.info(f"Running model.{subtask} ...")
                getattr(model, subtask)(**subtask_cfg)
            else:
                log.warning(f"{subtask} is not a valid subtask")

        log.info(f">>> Elapsed time: {elapsed()} <<< ")


def corpora_tasks(**cfg):
    args = eKonf.to_config(cfg)
    corpus_cfg = args.corpus
    pipeline = args.get("pipeline", {})
    if pipeline is None:
        raise Exception("Pipeline is missing")

    with elapsed_timer(format_time=True) as elapsed:
        corpora = eKonf.instantiate(corpus_cfg)
        corpora.load()
        corpora.concat_corpora()

        _pipeline_ = pipeline.get(eKonf.Keys.PIPELINE, {})
        df = apply_pipeline(corpora._data, _pipeline_, pipeline)
        log.info(f">>> Elapsed time: {elapsed()} <<< ")

    return df


def corpus_tasks(**cfg):
    from ekorpkit.corpora import Corpus

    args = eKonf.to_config(cfg)
    corpus_cfg = args.corpus
    merge_metadata = args.get("merge_metadata", False)
    pipeline = args.get("pipeline", {})
    if pipeline is None:
        raise Exception("Pipeline is missing")

    with elapsed_timer(format_time=True) as elapsed:
        corpora = eKonf.instantiate(corpus_cfg)
        if isinstance(corpora, Corpus):
            corpora = [corpora]

        for corpus in corpora:
            log.info(f"::: processing {corpus.name}")
            if merge_metadata:
                corpus.merge_metadata()
            update_args = {"corpus_name": corpus.name}
            _pipeline_ = pipeline.get(eKonf.Keys.PIPELINE, {})
            df = apply_pipeline(
                corpus._data, _pipeline_, pipeline, update_args=update_args
            )
        log.info(f">>> Elapsed time: {elapsed()} <<< ")

    return df


def transfomer_finetune(**cfg):
    args = eKonf.to_config(cfg)
    model_cfg = args.model.transformer.finetune
    dataset_cfg = args.dataset

    if model_cfg._target_:
        model = eKonf.instantiate(model_cfg, dataset_cfg=dataset_cfg)
        model.apply_pipeline()
    else:
        raise Exception("Model instantiation target is missing")


def dataframe_tasks(**cfg):
    args = eKonf.to_config(cfg)
    pipeline_args = args.get("pipeline", {})

    if pipeline_args._target_:
        df = eKonf.instantiate(pipeline_args)
        if df is None:
            raise Exception("No dataframe found")
    else:
        raise Exception("Dataframe instantiation target is missing")

    return df



import logging
from enum import Enum
from pathlib import Path
from pprint import pprint
from re import S
from ekorpkit import eKonf
from ekorpkit.pipelines.pipe import apply_pipeline
from ekorpkit.io.file import load_dataframe


log = logging.getLogger(__name__)


class _SPLITS(str, Enum):
    """Split keys in configs used by Dataset."""

    TRAIN = "train"
    DEV = "dev"
    TEST = "test"


class Dataset:
    """Dataset class."""

    SPLITS = _SPLITS

    def __init__(self, **args):
        self.args = eKonf.to_dict(args)
        self.name = self.args["name"]
        if isinstance(self.name, list):
            self.name = self.name[0]
        self.verbose = self.args.get("verbose", False)
        self.autoload = self.args.get("autoload", False)
        use_name_as_subdir = args.get("use_name_as_subdir", True)

        self.data_dir = Path(self.args["data_dir"])
        if use_name_as_subdir:
            self.data_dir = self.data_dir / self.name
        self.info_file = self.data_dir / f"info-{self.name}.yaml"
        self.info = eKonf.load(self.info_file) if self.info_file.is_file() else {}
        if self.info:
            if self.verbose:
                log.info(f"Loaded info file: {self.info_file}")
            self.args = eKonf.to_dict(eKonf.merge(self.args, self.info))
            self.info = eKonf.to_dict(self.info)

        if self.verbose:
            log.info(f"Intantiating a dataset {self.name} with a config:")
            pprint(eKonf.to_dict(self.args))

        self.filetype = self.args.get("filetype", "csv")
        self.data_files = self.args.get("data_files", None)
        if self.data_files is None:
            self.data_files = {
                _SPLITS.TRAIN: f"{self.name}-train.{self.filetype}",
                _SPLITS.DEV: f"{self.name}-dev.{self.filetype}",
                _SPLITS.TEST: f"{self.name}-test.{self.filetype}",
            }

        self.description = self.args.get("description", "")
        self.license = self.args.get("license", "")
        self._column_info = self.args.get("column_info")
        if self._column_info is None:
            raise ValueError("Column info can't be None")

        self._column = eKonf.instantiate(self._column_info)

        self.pipeline_args = self.args.get("pipeline", {})
        self.process_pipeline = self.pipeline_args.get(eKonf.Keys.PIPELINE, [])
        if self.process_pipeline is None:
            self.process_pipeline = []

        self.splits = {}
        self._loaded = False

        if self.autoload:
            self.load()

    def __str__(self):
        classname = self.__class__.__name__
        s = f"{classname} : {self.name}"
        return s

    def __getitem__(self, split):
        return self.splits[split]

    @property
    def COLUMN(self):
        return self._column

    @property
    def ID(self):
        return self.COLUMN.ID

    @property
    def IDs(self):
        return self.COLUMN.IDs

    @property
    def DATA(self):
        return self.COLUMN.DATA

    @property
    def DATATYPEs(self):
        return self.COLUMN.DATATYPEs

    def load(self):
        if self._loaded:
            return
        for split, data_file in self.data_files.items():
            data_file = self.data_dir / data_file
            df = load_dataframe(data_file, dtype=self.DATATYPEs)
            df = self.COLUMN.append_split(df, split)
            if self.process_pipeline and len(self.process_pipeline) > 0:
                df = apply_pipeline(df, self.process_pipeline, self.pipeline_args)
            self.splits[split] = df
        self._loaded = True



import logging
import platform
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib import font_manager, rc
from pathlib import Path


log = logging.getLogger(__name__)


def set_figure(
    ax,
    xlabel=None,
    ylabel=None,
    title=None,
    legend=None,
    grid=None,
    xlim=None,
    ylim=None,
    xticks=None,
    yticks=None,
    xticklabels=None,
    yticklabels=None,
    xtickmajorformatterfunc=None,
    ytickmajorformatterfunc=None,
    **kwargs,
):
    if xlabel is not None:
        if isinstance(xlabel, str):
            ax.set_xlabel(xlabel)
        else:
            ax.set_xlabel(**xlabel)
    if ylabel is not None:
        if isinstance(ylabel, str):
            ax.set_ylabel(ylabel)
        else:
            ax.set_ylabel(**ylabel)
    if title is not None:
        if isinstance(title, str):
            ax.set_title(title)
        else:
            ax.set_title(**title)
    if legend is not None:
        if isinstance(legend, (str, list)):
            ax.legend(legend)
        elif legend.get("labels") is not None:
            ax.legend(**legend)
    if grid is not None:
        if isinstance(grid, bool):
            ax.grid(grid)
        elif isinstance(grid, dict):
            ax.grid(**grid)
    if xticks is not None:
        if isinstance(xticks, str):
            ax.set_xticks(eval(xticks))
        else:
            if xticks.get("labels", None) or xticks.get("ticks", None):
                ax.set_xticks(**xticks)
    if yticks is not None:
        if isinstance(yticks, str):
            ax.set_yticks(eval(yticks))
        else:
            if yticks.get("labels", None) or yticks.get("ticks", None):
                ax.set_yticks(**yticks)
    if xticklabels is not None:
        if not xticklabels.get("labels", None):
            xticklabels["labels"] = ax.get_xticks().tolist()
        ax.set_xticklabels(**xticklabels)
    if yticklabels is not None:
        if not yticklabels.get("labels", None):
            yticklabels["labels"] = ax.get_yticks().tolist()
        ax.set_yticklabels(**yticklabels)
    if xlim is not None:
        if isinstance(xlim, str):
            ax.set_xlim(eval(xlim))
        else:
            ax.set_xlim(**xlim)
    if ylim is not None:
        if isinstance(ylim, str):
            ax.set_ylim(eval(ylim))
        else:
            ax.set_ylim(**ylim)
    if xtickmajorformatterfunc is not None:
        ax.xaxis.set_major_formatter(eval(xtickmajorformatterfunc))
    if ytickmajorformatterfunc is not None:
        ax.yaxis.set_major_formatter(eval(ytickmajorformatterfunc))


def set_style(style, rcParams, fontpath=None, language=None, **kwargs):
    if language or (fontpath and Path(fontpath).is_file()):
        fontname, fontpath = _configure_font(fontpath=fontpath)
        if fontname:
            rcParams["font.family"] = fontname

    plt.style.use(style)
    plt.rcParams.update(rcParams)


def _configure_font(set_font_for_matplot=True, fontpath=None, verbose=False):
    if not fontpath:
        if platform.system() == "Darwin":
            fontpath = "/System/Library/Fonts/Supplemental/AppleGothic.ttf"
        elif platform.system() == "Windows":
            fontpath = "c:/Windows/Fonts/malgun.ttf"
        elif platform.system() == "Linux":
            fontpath = "/usr/share/fonts/truetype/nanum/NanumGothic.ttf"
        if verbose:
            print("Font path:", fontpath)

    if not Path(fontpath).is_file():
        fontname = None
        fontpath = None
        print(f"Font file does not exist at {fontpath}")
        if platform.system() == "Linux":
            font_install_help = """
            apt install fontconfig
            apt install fonts-nanum
            fc-list | grep -i nanum
            """
            print(font_install_help)
    else:
        font_manager.fontManager.addfont(fontpath)
        fontname = font_manager.FontProperties(fname=fontpath).get_name()

        if set_font_for_matplot and fontname:
            rc("font", family=fontname)
            plt.rcParams["axes.unicode_minus"] = False
    if verbose:
        print("font family: ", plt.rcParams["font.family"])
        print("font path: ", fontpath)
    return fontname, fontpath



import os


def check_path(path):
    return os.path.exists(path)


def check_dir(filepath):
    dirname = os.path.abspath(os.path.dirname(filepath))
    if not os.path.exists(dirname):
        os.makedirs(dirname)



import logging
import codecs
import pandas as pd
from pathlib import Path
from ekorpkit import eKonf
from ekorpkit.io.file import load_dataframe, get_filepaths

DESCRIPTION = "Corpus for Language Models"
LICENSE = "Copyright of the corpus is owned by the authors."


log = logging.getLogger(__name__)


class Corpus:
    def __init__(self, **args):
        self.args = eKonf.to_dict(args)
        self.name = self.args["name"]
        if isinstance(self.name, list):
            self.name = self.name[0]
        self.verbose = self.args.get("verbose", False)
        self.autoload = self.args.get("autoload", False)
        self.automerge = self.args.get("automerge", False)
        use_name_as_subdir = args.get("use_name_as_subdir", True)

        self.data_dir = Path(self.args["data_dir"])
        if use_name_as_subdir:
            self.data_dir = self.data_dir / self.name
        self.metadata_dir = self.args.get("metadata_dir", None)
        if self.metadata_dir is None:
            self.metadata_dir = self.data_dir
        else:
            self.metadata_dir = Path(self.metadata_dir)
            if use_name_as_subdir:
                self.metadata_dir = self.metadata_dir / self.name
        self.info_file = self.data_dir / f"info-{self.name}.yaml"
        self.info = eKonf.load(self.info_file) if self.info_file.is_file() else {}
        if self.info:
            if self.verbose:
                log.info(f"Loaded info file: {self.info_file}")
            self.args = eKonf.to_dict(eKonf.update(self.args, self.info))
            self.info = eKonf.to_dict(self.info)

        if self.verbose:
            log.info(f"Intantiating a corpus {self.name} with a config:")
            eKonf.pprint(self.args)

        self.filetype = self.args.get("filetype", "csv")
        self.data_files = self.args.get("data_files", None)
        self.meta_files = self.args.get("meta_files", None)
        if self.data_files is None:
            self.data_files = {
                "train": f"{self.name}*{self.filetype}*",
            }

        self._collapse_ids = self.args.get("collapse_ids", False)
        self.description = self.args.get("description", DESCRIPTION)
        self.license = self.args.get("license", LICENSE)
        self.split_info = self.args.get("splits", None)
        self._column_info = self.args.get("column_info", None)
        if self._column_info is None:
            raise ValueError("Column info can't be None")

        self._column = eKonf.instantiate(self._column_info)

        self._data = None
        self._metadata = None
        self._metadata_merged = False
        self._loaded = False

        if self.autoload:
            self.load()
            self.load_metadata()
            self.load_timestamp()
            if self.automerge:
                self.merge_metadata()

    def __str__(self):
        classname = self.__class__.__name__
        s = f"{classname} : {self.name}"
        return s

    @property
    def COLUMN(self):
        return self._column

    @property
    def ID(self):
        return self.COLUMN.ID

    @property
    def IDs(self):
        return self.COLUMN.IDs

    @property
    def TEXT(self):
        return self.COLUMN.TEXT

    @property
    def DATA(self):
        return self.COLUMN.DATA

    @property
    def METADATA(self):
        return self.COLUMN.METADATA

    @property
    def data(self):
        return self._data

    @property
    def metadata(self):
        return self._metadata

    def __len__(self):
        return len(self._data)

    @property
    def num_rows(self) -> int:
        """Number of rows in the corpus (same as :meth:`Corpus.__len__`)."""
        if self._data.index is not None:
            return len(self._data.index)
        return len(self._data)

    def load_timestamp(self):
        self._data, self._metadata = self.COLUMN.to_timestamp(
            self._data, self._metadata
        )
        if self.verbose:
            print(self._data.head())

    def load(self):
        dfs = []
        for split, data_file in self.data_files.items():
            filepaths = get_filepaths(data_file, self.data_dir)
            df = pd.concat([load_dataframe(f, verbose=self.verbose) for f in filepaths])
            df = self.COLUMN.combine_texts(df)
            if self._collapse_ids:
                df = self.COLUMN.append_split(df, split)
            dfs.append(df[self.COLUMN.DATA])
        self._data = pd.concat(dfs)
        if self._collapse_ids:
            self.COLUMN.combine_ids(self._data)
        if self.verbose:
            log.info(f"Data loaded {len(self._data)} rows")
            print(self._data.head(3))
            print(self._data.tail(3))
        self._loaded = True

    def load_metadata(self):
        if not self.meta_files or len(self.meta_files) == 0:
            log.info("No metadata files found")
            return

        dfs = []
        for split, data_file in self.meta_files.items():
            filepaths = get_filepaths(data_file, self.metadata_dir)
            df = pd.concat([load_dataframe(f, verbose=self.verbose) for f in filepaths])
            if self._collapse_ids:
                df = self.COLUMN.append_split(df, split)
            dfs.append(df)
        self._metadata = pd.concat(dfs)
        if self._collapse_ids:
            self.COLUMN.combine_ids(self._metadata)
        if self.verbose:
            log.info(f"Metadata loaded {len(self._metadata)} rows")
            print(self._metadata.head(3))
            print(self._metadata.tail(3))

    def merge_metadata(self):
        if self._metadata is None or self._metadata_merged:
            return
        self._data = self.COLUMN.merge_metadata(self._data, self._metadata)
        self._metadata_merged = True
        if self.verbose:
            log.info(f"Metadata merged to data")
            print(self._data.head(3))
            print(self._data.tail(3))
